# pylint: disable=line-too-long,useless-suppression
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from azure.identity import DefaultAzureCredential

from azure.mgmt.applicationinsights import ApplicationInsightsManagementClient

"""
# PREREQUISITES
    pip install azure-identity
    pip install azure-mgmt-applicationinsights
# USAGE
    python workbook_template_add.py

    Before run the sample, please set the values of the client ID, tenant ID and client secret
    of the AAD application as environment variables: AZURE_CLIENT_ID, AZURE_TENANT_ID,
    AZURE_CLIENT_SECRET. For more info about how to get the value, please see:
    https://docs.microsoft.com/azure/active-directory/develop/howto-create-service-principal-portal
"""


def main():
    client = ApplicationInsightsManagementClient(
        credential=DefaultAzureCredential(),
        subscription_id="SUBSCRIPTION_ID",
    )

    response = client.workbook_templates.create_or_update(
        resource_group_name="my-resource-group",
        resource_name="testtemplate2",
        workbook_template_properties={
            "location": "west us",
            "properties": {
                "author": "Contoso",
                "galleries": [
                    {
                        "category": "Failures",
                        "name": "Simple Template",
                        "order": 100,
                        "resourceType": "microsoft.insights/components",
                        "type": "tsg",
                    }
                ],
                "priority": 1,
                "templateData": {
                    "$schema": "https://github.com/Microsoft/Application-Insights-Workbooks/blob/master/schema/workbook.json",
                    "items": [
                        {
                            "content": {
                                "json": "## New workbook\n---\n\nWelcome to your new workbook.  This area will display text formatted as markdown.\n\n\nWe've included a basic analytics query to get you started. Use the `Edit` button below each section to configure it or add more sections."
                            },
                            "name": "text - 2",
                            "type": 1,
                        },
                        {
                            "content": {
                                "exportToExcelOptions": "visible",
                                "query": "union withsource=TableName *\n| summarize Count=count() by TableName\n| render barchart",
                                "queryType": 0,
                                "resourceType": "microsoft.operationalinsights/workspaces",
                                "size": 1,
                                "version": "KqlItem/1.0",
                            },
                            "name": "query - 2",
                            "type": 3,
                        },
                    ],
                    "styleSettings": {},
                    "version": "Notebook/1.0",
                },
            },
            "tags": None,
        },
    )
    print(response)


# x-ms-original-file: 2020-11-20/WorkbookTemplateAdd.json
if __name__ == "__main__":
    main()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.
import unittest
import uuid

import asyncio

import pytest
from typing import Dict, Any, Optional

import test_config
from azure.core.pipeline.transport._aiohttp import AioHttpTransport
from azure.core.exceptions import ServiceResponseError
from azure.cosmos.exceptions import CosmosHttpResponseError
from azure.cosmos.aio import CosmosClient
from _fault_injection_transport import FaultInjectionTransport
from _fault_injection_transport_async import FaultInjectionTransportAsync
from test_per_partition_automatic_failover import create_failover_errors, create_threshold_errors, session_retry_hook, ppaf_user_agent_hook
from test_per_partition_circuit_breaker_mm import REGION_1, REGION_2, PK_VALUE, BATCH, write_operations_errors_and_boolean
from test_per_partition_circuit_breaker_mm_async import perform_write_operation

#cspell:ignore PPAF, ppaf

# These tests assume that the configured live account has one main write region and one secondary read region.

@pytest.mark.cosmosPerPartitionAutomaticFailover
@pytest.mark.cosmosAADPerPartitionAutomaticFailover
@pytest.mark.asyncio
class TestPerPartitionAutomaticFailoverAsync:
    host = test_config.TestConfig.host
    master_key = test_config.TestConfig.masterKey
    TEST_DATABASE_ID = test_config.TestConfig.TEST_DATABASE_ID
    TEST_CONTAINER_MULTI_PARTITION_ID = test_config.TestConfig.TEST_MULTI_PARTITION_CONTAINER_ID

    async def setup_method_with_custom_transport(self, custom_transport: Optional[AioHttpTransport],
                                                 default_endpoint=None, read_first=False, **kwargs):
        regions = [REGION_2, REGION_1] if read_first else [REGION_1, REGION_2]
        endpoint = default_endpoint or self.host
        container_id = kwargs.pop("container_id", None)
        exclude_client_regions = kwargs.pop("exclude_client_regions", False)
        excluded_regions = []
        if exclude_client_regions:
            excluded_regions = [REGION_2]
        if not container_id:
            container_id = self.TEST_CONTAINER_MULTI_PARTITION_ID
        client_kwargs = {
            "consistency_level": "Session",
            "preferred_locations": regions,
            "excluded_locations": excluded_regions,
            "transport": custom_transport,
            **kwargs,
        }
        client = test_config.TestConfig.create_data_client_async_for_endpoint(endpoint, **client_kwargs)
        db = client.get_database_client(self.TEST_DATABASE_ID)
        container = db.get_container_client(container_id)
        await client.__aenter__()
        return {"client": client, "db": db, "col": container}
    
    @staticmethod
    async def cleanup_method(initialized_objects: Dict[str, Any]):
        method_client: CosmosClient = initialized_objects["client"]
        await method_client.close()

    async def setup_info(self, error=None, max_count=None, is_batch=False, exclude_client_regions=False, session_error=False, **kwargs):
        custom_transport = FaultInjectionTransportAsync()
        # two documents targeted to same partition, one will always fail and the other will succeed
        doc_fail_id = str(uuid.uuid4())
        doc_success_id = str(uuid.uuid4())
        predicate = lambda r: (FaultInjectionTransportAsync.predicate_req_for_document_with_id(r, doc_fail_id) and
                               FaultInjectionTransportAsync.predicate_is_write_operation(r, "documents"))
        # The MockRequest only gets used to create the MockHttpResponse
        mock_request = FaultInjectionTransport.MockHttpRequest(url=self.host)
        if is_batch:
            success_response = FaultInjectionTransportAsync.MockHttpResponse(mock_request, 200, [{"statusCode": 200}],)
        else:
            success_response = FaultInjectionTransportAsync.MockHttpResponse(mock_request, 200)
        if error:
            custom_transport.add_fault(predicate=predicate, fault_factory=error, max_inner_count=max_count,
                                       after_max_count=success_response)
        if session_error:
            read_predicate = lambda r: (FaultInjectionTransportAsync.predicate_is_operation_type(r, "Read")
                                        and FaultInjectionTransportAsync.predicate_req_for_document_with_id(r, doc_fail_id))
            read_error = CosmosHttpResponseError(
                            status_code=404,
                            message="Some injected error.",
                            sub_status=1002)
            error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(0, read_error))
            success_response = FaultInjectionTransportAsync.MockHttpResponse(mock_request, 200)
            custom_transport.add_fault(predicate=read_predicate, fault_factory=error_lambda, max_inner_count=max_count,
                                       after_max_count=success_response)
        is_get_account_predicate = lambda r: FaultInjectionTransportAsync.predicate_is_database_account_call(r)
        # Set the database account response to have PPAF enabled
        ppaf_enabled_database_account = \
            lambda r, inner: FaultInjectionTransportAsync.transform_topology_ppaf_enabled(inner=inner)
        custom_transport.add_response_transformation(
            is_get_account_predicate,
            ppaf_enabled_database_account)
        setup = await self.setup_method_with_custom_transport(None, default_endpoint=self.host,
                                                        exclude_client_regions=exclude_client_regions, **kwargs)
        custom_setup = await self.setup_method_with_custom_transport(custom_transport, default_endpoint=self.host,
                                                        exclude_client_regions=exclude_client_regions, **kwargs)
        return setup, doc_fail_id, doc_success_id, custom_setup, custom_transport, predicate

    @pytest.mark.parametrize("write_operation, error, exclude_regions", write_operations_errors_and_boolean(create_failover_errors()))
    async def test_ppaf_partition_info_cache_and_routing_async(self, write_operation, error, exclude_regions):
        # This test validates that the partition info cache is updated correctly upon failures, and that the
        # per-partition automatic failover logic routes requests to the next available regional endpoint.
        # We also verify that this logic is unaffected by user excluded regions, since write-region routing is
        # entirely taken care of on the service.
        error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(0, error))
        setup, doc_fail_id, doc_success_id, custom_setup, custom_transport, predicate = await self.setup_info(error_lambda, 1,
                                                                                                              write_operation == BATCH, exclude_client_regions=exclude_regions)
        try:
            container = setup['col']
            fault_injection_container = custom_setup['col']
            global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager

            # Create a document to populate the per-partition GEM partition range info cache
            await fault_injection_container.create_item(body={'id': doc_success_id, 'pk': PK_VALUE,
                                                        'name': 'sample document', 'key': 'value'})
            pk_range_wrapper = list(global_endpoint_manager.partition_range_to_failover_info.keys())[0]
            initial_region = global_endpoint_manager.partition_range_to_failover_info[pk_range_wrapper].current_region

            # Based on our configuration, we should have had one error followed by a success - marking only the previous endpoint as unavailable
            await perform_write_operation(
                write_operation,
                container,
                fault_injection_container,
                doc_fail_id,
                PK_VALUE)
            partition_info = global_endpoint_manager.partition_range_to_failover_info[pk_range_wrapper]
            # Verify that the partition is marked as unavailable, and that the current regional endpoint is not the same
            assert len(partition_info.unavailable_regional_endpoints) == 1
            assert initial_region in partition_info.unavailable_regional_endpoints
            assert initial_region != partition_info.current_region # west us 3 != west us

            # Now we run another request to see how the cache gets updated
            await perform_write_operation(
                write_operation,
                container,
                fault_injection_container,
                doc_fail_id,
                PK_VALUE)
            partition_info = global_endpoint_manager.partition_range_to_failover_info[pk_range_wrapper]
            # Verify that the cache is empty, since the request going to the second regional endpoint failed
            # Once we reach the point of all available regions being marked as unavailable, the cache is cleared
            assert len(partition_info.unavailable_regional_endpoints) == 0
            assert initial_region not in partition_info.unavailable_regional_endpoints
            assert partition_info.current_region is None
        finally:
            await self.cleanup_method(custom_setup)
            await self.cleanup_method(setup)

    @pytest.mark.parametrize("write_operation, error, exclude_regions", write_operations_errors_and_boolean(create_threshold_errors()))
    async def test_ppaf_partition_thresholds_and_routing_async(self, write_operation, error, exclude_regions):
        # This test validates that the partition info cache is updated correctly upon failures, and that the
        # per-partition automatic failover logic routes requests to the next available regional endpoint.
        # We also verify that this logic is unaffected by user excluded regions, since write-region routing is
        # entirely taken care of on the service.
        error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(0, error))
        setup, doc_fail_id, doc_success_id, custom_setup, custom_transport, predicate = await self.setup_info(error_lambda,
                                                                                                              exclude_client_regions=exclude_regions,)
        try:
            container = setup['col']
            fault_injection_container = custom_setup['col']
            global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager

            # Create a document to populate the per-partition GEM partition range info cache
            await fault_injection_container.create_item(body={'id': doc_success_id, 'pk': PK_VALUE,
                                                        'name': 'sample document', 'key': 'value'})
            pk_range_wrapper = list(global_endpoint_manager.partition_range_to_failover_info.keys())[0]
            initial_region = global_endpoint_manager.partition_range_to_failover_info[pk_range_wrapper].current_region

            consecutive_failures = 6
            for i in range(consecutive_failures):
                # We perform the write operation multiple times to check the consecutive failures logic
                with pytest.raises((CosmosHttpResponseError, ServiceResponseError)) as exc_info:
                    await perform_write_operation(write_operation,
                                                  container,
                                                  fault_injection_container,
                                                  doc_fail_id,
                                                  PK_VALUE)
                assert exc_info.value == error

            # Verify that the threshold for consecutive failures is updated
            pk_range_wrappers = list(global_endpoint_manager.ppaf_thresholds_tracker.pk_range_wrapper_to_failure_count.keys())
            assert len(pk_range_wrappers) == 1
            failure_count = global_endpoint_manager.ppaf_thresholds_tracker.pk_range_wrapper_to_failure_count[pk_range_wrappers[0]]
            assert failure_count == consecutive_failures

            # Verify that a single success to the same partition resets the consecutive failures count
            await perform_write_operation(write_operation,
                                    container,
                                    fault_injection_container,
                                    str(uuid.uuid4()),
                                    PK_VALUE)

            failure_count = global_endpoint_manager.ppaf_thresholds_tracker.pk_range_wrapper_to_failure_count.get(pk_range_wrappers[0], 0)
            assert failure_count == 0

            # Run enough failed requests to the partition to trigger the failover logic
            for i in range(12):
                with pytest.raises((CosmosHttpResponseError, ServiceResponseError)) as exc_info:
                    await perform_write_operation(write_operation,
                                                  container,
                                                  fault_injection_container,
                                                  doc_fail_id,
                                                  PK_VALUE)
                assert exc_info.value == error
            # We should have marked the previous endpoint as unavailable after 10 successive failures
            partition_info = global_endpoint_manager.partition_range_to_failover_info[pk_range_wrapper]
            # Verify that the partition is marked as unavailable, and that the current regional endpoint is not the same
            assert len(partition_info.unavailable_regional_endpoints) == 1
            assert initial_region in partition_info.unavailable_regional_endpoints
            assert initial_region != partition_info.current_region # west us 3 != west us

            # 12 failures - 10 to trigger failover, 2 more to start counting again
            failure_count = global_endpoint_manager.ppaf_thresholds_tracker.pk_range_wrapper_to_failure_count[pk_range_wrappers[0]]
            assert failure_count == 2
        finally:
            await self.cleanup_method(custom_setup)
            await self.cleanup_method(setup)

    @pytest.mark.parametrize("write_operation, error, exclude_regions", write_operations_errors_and_boolean(create_failover_errors()))
    async def test_ppaf_session_unavailable_retry_async(self, write_operation, error, exclude_regions):
        # Account config has 2 regions: West US 3 (A) and West US (B). This test validates that after marking the write
        # region (A) as unavailable, the next request is retried to the read region (B) and succeeds. The next read request
        # should see that the write region (A) is unavailable for the partition, and should retry to the read region (B) as well.
        # We also verify that this logic is unaffected by user excluded regions, since write-region routing is
        # entirely taken care of on the service.
        error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(0, error))
        setup, doc_fail_id, doc_success_id, custom_setup, custom_transport, predicate = await self.setup_info(error_lambda, max_count=1,
                                                                                                        is_batch=write_operation==BATCH,
                                                                                                        session_error=True, exclude_client_regions=exclude_regions)
        try:
            container = setup['col']
            fault_injection_container = custom_setup['col']
            global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager

            # Create a document to populate the per-partition GEM partition range info cache
            await fault_injection_container.create_item(body={'id': doc_success_id, 'pk': PK_VALUE,
                                                        'name': 'sample document', 'key': 'value'})
            pk_range_wrapper = list(global_endpoint_manager.partition_range_to_failover_info.keys())[0]
            initial_region = global_endpoint_manager.partition_range_to_failover_info[pk_range_wrapper].current_region

            # Verify the region that is being used for the read requests
            read_response = await fault_injection_container.read_item(doc_success_id, PK_VALUE)
            uri = read_response.get_response_headers().get('Content-Location')
            region = fault_injection_container.client_connection._global_endpoint_manager.location_cache.get_location_from_endpoint(uri)
            assert region == REGION_1 # first preferred region

            # Based on our configuration, we should have had one error followed by a success - marking only the previous endpoint as unavailable
            await perform_write_operation(
                    write_operation,
                    container,
                    fault_injection_container,
                    doc_fail_id,
                    PK_VALUE)
            partition_info = global_endpoint_manager.partition_range_to_failover_info[pk_range_wrapper]
            # Verify that the partition is marked as unavailable, and that the current regional endpoint is not the same
            assert len(partition_info.unavailable_regional_endpoints) == 1
            assert initial_region in partition_info.unavailable_regional_endpoints
            assert initial_region != partition_info.current_region # west us 3 != west us

            # Now we run a read request that runs into a 404.1002 error, which should retry to the read region
            # We verify that the read request was going to the correct region by using the raw_response_hook
            await fault_injection_container.read_item(doc_fail_id, PK_VALUE, raw_response_hook=session_retry_hook)
        finally:
            await self.cleanup_method(custom_setup)
            await self.cleanup_method(setup)

    async def test_ppaf_user_agent_feature_flag_async(self):
        # Simple test to verify the user agent suffix is being updated with the relevant feature flags
        setup, doc_fail_id, doc_success_id, custom_setup, custom_transport, predicate = await self.setup_info()
        try:
            fault_injection_container = custom_setup['col']
            # Create a document to check the response headers
            await fault_injection_container.upsert_item(body={'id': doc_success_id, 'pk': PK_VALUE, 'name': 'sample document', 'key': 'value'},
                                                        raw_response_hook=ppaf_user_agent_hook)
        finally:
            await self.cleanup_method(custom_setup)
            await self.cleanup_method(setup)

if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Performance tests for the breaking_changes_checker.

These tests demonstrate and verify the performance improvements made to:
1. run_async_cleanup() - O(n) set-based approach vs O(n*m) nested loop + list.remove()
2. get_reportable_changes() - single-pass filtering vs deepcopy + list.remove()
3. AST caching - _get_parsed_module() and _find_class_node() helpers
4. ClassTreeAnalyzer early exit
"""

import time
from breaking_changes_checker.breaking_changes_tracker import BreakingChangesTracker, BreakingChangeType
from breaking_changes_checker.changelog_tracker import ChangelogTracker


def _generate_large_stable_current(num_modules, num_classes_per_module, num_methods_per_class, num_params_per_method):
    """Generate large stable/current API reports for performance testing."""
    stable = {}
    current = {}
    for m in range(num_modules):
        module_name = f"azure.mgmt.network.v2023_01_{m:02d}"
        stable[module_name] = {"class_nodes": {}, "function_nodes": {}}
        current[module_name] = {"class_nodes": {}, "function_nodes": {}}
        for c in range(num_classes_per_module):
            class_name = f"NetworkClass{c}"
            methods_stable = {}
            methods_current = {}
            for me in range(num_methods_per_class):
                method_name = f"method_{me}"
                params = {}
                for p in range(num_params_per_method):
                    params[f"param_{p}"] = {
                        "default": None,
                        "param_type": "positional_or_keyword"
                    }
                methods_stable[method_name] = {
                    "parameters": dict(params),
                    "is_async": False,
                }
                methods_current[method_name] = {
                    "parameters": dict(params),
                    "is_async": False,
                }
            stable[module_name]["class_nodes"][class_name] = {
                "type": None,
                "methods": methods_stable,
                "properties": {f"prop_{i}": {"attr_type": "str"} for i in range(3)}
            }
            current[module_name]["class_nodes"][class_name] = {
                "type": None,
                "methods": methods_current,
                "properties": {f"prop_{i}": {"attr_type": "str"} for i in range(3)}
            }
    return stable, current


def _generate_changes_list(n, include_aio=True):
    """Generate a list of n breaking changes, half sync and half aio if include_aio."""
    changes = []
    for i in range(n):
        module = f"azure.mgmt.network"
        if include_aio and i % 2 == 1:
            module = f"azure.mgmt.network.aio"
        changes.append((
            "Deleted or renamed model `{}`",
            BreakingChangeType.REMOVED_OR_RENAMED_CLASS,
            module, f"SomeClass{i // 2 if include_aio else i}"
        ))
    return changes


class TestAsyncCleanupPerformance:
    """Test performance of the optimized run_async_cleanup method."""

    def test_async_cleanup_correctness_small(self):
        """Verify correctness of optimized async cleanup with a small dataset."""
        changes = [
            ("msg", BreakingChangeType.REMOVED_OR_RENAMED_CLASS, "azure.mgmt.network", "ClassA"),
            ("msg", BreakingChangeType.REMOVED_OR_RENAMED_CLASS, "azure.mgmt.network.aio", "ClassA"),
            ("msg", BreakingChangeType.REMOVED_OR_RENAMED_CLASS, "azure.mgmt.network", "ClassB"),
            ("msg", BreakingChangeType.REMOVED_OR_RENAMED_CLASS, "azure.mgmt.network.aio", "ClassC"),  # no sync counterpart
        ]
        stable = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        current = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        tracker = BreakingChangesTracker(stable, current, "azure-mgmt-network")
        tracker.run_async_cleanup(changes)

        # aio ClassA should be removed (has sync counterpart)
        # aio ClassC should remain (no sync counterpart)
        assert len(changes) == 3
        modules = [c[2] for c in changes]
        assert "azure.mgmt.network.aio" in modules  # ClassC remains
        class_names = [c[3] for c in changes]
        assert "ClassA" in class_names  # sync ClassA remains
        assert "ClassB" in class_names  # sync ClassB remains
        assert "ClassC" in class_names  # aio ClassC remains (no sync counterpart)

    def test_async_cleanup_performance_large(self):
        """Benchmark async cleanup with a large number of changes.
        
        The old O(n*m) implementation with list.remove() would be very slow
        for large change lists. The new set-based O(n) approach should handle
        this in milliseconds.
        """
        n = 10000  # 10,000 changes (5,000 sync + 5,000 aio)
        changes = _generate_changes_list(n, include_aio=True)
        stable = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        current = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        tracker = BreakingChangesTracker(stable, current, "azure-mgmt-network")

        start = time.perf_counter()
        tracker.run_async_cleanup(changes)
        elapsed = time.perf_counter() - start

        # Should complete in well under 1 second with the optimized implementation
        # The old implementation would take 10+ seconds for this size
        assert elapsed < 1.0, f"run_async_cleanup took {elapsed:.3f}s for {n} changes (expected < 1s)"
        # 5000 aio changes with sync counterparts should be removed
        assert len(changes) == n // 2
        # All remaining should be sync changes
        assert all("aio" not in c[2] for c in changes)

    def test_async_cleanup_performance_very_large(self):
        """Benchmark with 50,000 changes to show scalability."""
        n = 50000
        changes = _generate_changes_list(n, include_aio=True)
        stable = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        current = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        tracker = BreakingChangesTracker(stable, current, "azure-mgmt-network")

        start = time.perf_counter()
        tracker.run_async_cleanup(changes)
        elapsed = time.perf_counter() - start

        assert elapsed < 2.0, f"run_async_cleanup took {elapsed:.3f}s for {n} changes (expected < 2s)"
        assert len(changes) == n // 2

    def test_async_cleanup_with_list_args(self):
        """Test that async cleanup handles unhashable types (lists) in change tuples."""
        changes = [
            ("msg", BreakingChangeType.CHANGED_PARAMETER_ORDERING, "azure.mgmt.network",
             "MyClient", "my_method", ["param_a", "param_b"], ["param_b", "param_a"]),
            ("msg", BreakingChangeType.CHANGED_PARAMETER_ORDERING, "azure.mgmt.network.aio",
             "MyClient", "my_method", ["param_a", "param_b"], ["param_b", "param_a"]),
        ]
        stable = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        current = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        tracker = BreakingChangesTracker(stable, current, "azure-mgmt-network")
        tracker.run_async_cleanup(changes)

        # The aio version should be removed since it has a sync counterpart
        assert len(changes) == 1
        assert "aio" not in changes[0][2]


class TestReportableChangesPerformance:
    """Test performance of the optimized get_reportable_changes method."""

    def test_reportable_changes_no_ignores(self):
        """Verify that with no ignore rules, all changes are kept."""
        changes = _generate_changes_list(100, include_aio=False)
        stable = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        current = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        tracker = BreakingChangesTracker(stable, current, "azure-mgmt-network")
        tracker.get_reportable_changes({}, changes)
        assert len(changes) == 100

    def test_reportable_changes_performance_large(self):
        """Benchmark get_reportable_changes with many changes and ignore rules.
        
        The old implementation used deepcopy + list.remove() which is O(n^2).
        The new single-pass filtering is O(n*m) where m = number of ignore rules.
        """
        n = 5000
        changes = _generate_changes_list(n, include_aio=False)
        ignore_rules = {
            "azure-mgmt-.*": [
                (BreakingChangeType.REMOVED_OR_RENAMED_CLASS, "*", f"SomeClass{i}", None, None)
                for i in range(100)  # 100 ignore rules
            ]
        }
        stable = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        current = {"azure.mgmt.network": {"class_nodes": {}, "function_nodes": {}}}
        tracker = BreakingChangesTracker(stable, current, "azure-mgmt-network")

        start = time.perf_counter()
        tracker.get_reportable_changes(ignore_rules, changes)
        elapsed = time.perf_counter() - start

        # Should complete well under 1 second
        assert elapsed < 1.0, f"get_reportable_changes took {elapsed:.3f}s (expected < 1s)"
        # 100 of the 5000 changes should be filtered out
        assert len(changes) == n - 100


class TestCheckerRunPerformance:
    """Test full checker run performance with large API surfaces."""

    def test_full_checker_run_large_unchanged_api(self):
        """Benchmark a full checker run with a large unchanged API surface.
        
        When stable and current are identical, jsondiff produces an empty diff,
        so diff-based checks do minimal work. However, run_checks() still executes
        checks like parameter ordering across the full API surface. This test
        ensures the overall run remains fast and bounded even for a large, unchanged API.
        """
        stable, current = _generate_large_stable_current(
            num_modules=10,
            num_classes_per_module=50,
            num_methods_per_class=10,
            num_params_per_method=5,
        )

        start = time.perf_counter()
        tracker = BreakingChangesTracker(stable, current, "azure-mgmt-network")
        tracker.run_checks()
        elapsed = time.perf_counter() - start

        assert elapsed < 5.0, f"Full checker run took {elapsed:.3f}s for large unchanged API (expected < 5s)"
        assert len(tracker.breaking_changes) == 0

    def test_changelog_tracker_run_large_with_additions(self):
        """Benchmark ChangelogTracker with a large API that has many additions.
        
        This simulates a scenario similar to azure-mgmt-network where many new
        classes and methods are added.
        """
        stable, current = _generate_large_stable_current(
            num_modules=5,
            num_classes_per_module=20,
            num_methods_per_class=5,
            num_params_per_method=3,
        )
        # Add new classes and methods in current
        for m in range(5):
            module_name = f"azure.mgmt.network.v2023_01_{m:02d}"
            for c in range(10):
                new_class = f"NewNetworkClass{m}_{c}"
                current[module_name]["class_nodes"][new_class] = {
                    "type": None,
                    "methods": {
                        f"new_method_{me}": {
                            "parameters": {
                                f"param_{p}": {"default": None, "param_type": "positional_or_keyword"}
                                for p in range(3)
                            },
                            "is_async": False,
                        }
                        for me in range(5)
                    },
                    "properties": {f"prop_{i}": {"attr_type": "str"} for i in range(3)}
                }

        start = time.perf_counter()
        tracker = ChangelogTracker(stable, current, "azure-mgmt-network")
        tracker.run_checks()
        elapsed = time.perf_counter() - start

        assert elapsed < 5.0, f"ChangelogTracker run took {elapsed:.3f}s (expected < 5s)"
        assert len(tracker.features_added) > 0

    def test_parameter_ordering_check_large(self):
        """Benchmark check_parameter_ordering with many modules/classes/methods."""
        stable, current = _generate_large_stable_current(
            num_modules=10,
            num_classes_per_module=50,
            num_methods_per_class=10,
            num_params_per_method=5,
        )

        tracker = BreakingChangesTracker(stable, current, "azure-mgmt-network")

        start = time.perf_counter()
        tracker.check_parameter_ordering()
        elapsed = time.perf_counter() - start

        # 10 modules * 50 classes * 10 methods = 5000 method comparisons
        assert elapsed < 2.0, f"check_parameter_ordering took {elapsed:.3f}s (expected < 2s)"
        assert len(tracker.breaking_changes) == 0


class TestASTCachePerformance:
    """Test performance of the AST caching mechanism."""

    def test_ast_cache_avoids_redundant_parsing(self):
        """Verify that the AST cache is populated and reused."""
        from breaking_changes_checker.detect_breaking_changes import (
            _ast_cache, _get_parsed_module, _find_class_node
        )
        import tempfile
        import os

        # Create a temporary Python file with multiple classes
        with tempfile.NamedTemporaryFile(mode='w', suffix='.py', delete=False) as f:
            f.write(
                "class Foo:\n"
                "    x: int = 0\n"
                "\n"
                "class Bar:\n"
                "    y: str = ''\n"
                "\n"
                "class Baz:\n"
                "    z: float = 0.0\n"
            )
            temp_path = f.name

        try:
            _ast_cache.clear()

            # First call should parse and cache
            module1 = _get_parsed_module(temp_path)
            assert temp_path in _ast_cache

            # Second call should return cached result (same object)
            module2 = _get_parsed_module(temp_path)
            assert module1 is module2

            # Find multiple classes from the same cached module
            foo_node = _find_class_node(module1, "Foo")
            assert foo_node is not None
            assert foo_node.name == "Foo"

            bar_node = _find_class_node(module1, "Bar")
            assert bar_node is not None
            assert bar_node.name == "Bar"

            baz_node = _find_class_node(module1, "Baz")
            assert baz_node is not None
            assert baz_node.name == "Baz"
        finally:
            os.unlink(temp_path)
            _ast_cache.clear()

    def test_ast_cache_performance_many_lookups(self):
        """Benchmark AST cache with many lookups to the same file.
        
        Simulates the scenario where many classes in a large package
        share the same source file (e.g., _models.py in azure-mgmt-network).
        """
        from breaking_changes_checker.detect_breaking_changes import (
            _ast_cache, _get_parsed_module
        )
        import tempfile
        import os

        # Create a temporary file with many classes (simulating a large _models.py)
        num_classes = 200
        lines = []
        for i in range(num_classes):
            lines.append(f"class Model{i}:\n    attr_{i}: int = {i}\n\n")

        with tempfile.NamedTemporaryFile(mode='w', suffix='.py', delete=False) as f:
            f.write("".join(lines))
            temp_path = f.name

        try:
            _ast_cache.clear()

            # Without caching, 200 lookups would mean 200 file reads + 200 AST parses
            # With caching, it's 1 file read + 1 AST parse + 200 cache hits
            start = time.perf_counter()
            for i in range(num_classes):
                module = _get_parsed_module(temp_path)
            elapsed = time.perf_counter() - start

            # Should be nearly instantaneous since all after the first are cache hits
            assert elapsed < 0.1, f"{num_classes} cached lookups took {elapsed:.3f}s (expected < 0.1s)"
            assert len(_ast_cache) == 1  # Only one file cached
        finally:
            os.unlink(temp_path)
            _ast_cache.clear()


class TestClassTreeAnalyzerEarlyExit:
    """Test that ClassTreeAnalyzer exits early after finding the target class."""

    def test_early_exit_finds_class(self):
        """Verify early exit still finds the class correctly."""
        from breaking_changes_checker.detect_breaking_changes import _find_class_node
        import ast

        source = """
class First:
    pass

class Target:
    x: int = 0

class Third:
    pass
"""
        module = ast.parse(source)
        cls_node = _find_class_node(module, "Target")
        assert cls_node is not None
        assert cls_node.name == "Target"

    def test_early_exit_returns_none_for_missing(self):
        """Verify that early exit correctly returns None when class is not found."""
        from breaking_changes_checker.detect_breaking_changes import _find_class_node
        import ast

        source = """
class First:
    pass

class Second:
    pass
"""
        module = ast.parse(source)
        cls_node = _find_class_node(module, "NonExistent")
        assert cls_node is None

    def test_early_exit_performance(self):
        """Benchmark early exit vs full traversal for large source files.
        
        With early exit, finding a class near the top of a large file should
        be significantly faster than traversing the entire AST.
        """
        from breaking_changes_checker.detect_breaking_changes import _find_class_node
        import ast

        # Generate a large source file with many classes
        num_classes = 500
        lines = ["class TargetClass:\n    x: int = 0\n\n"]  # Target is first
        for i in range(num_classes):
            lines.append(f"class OtherClass{i}:\n    attr: int = {i}\n\n")
        source = "".join(lines)
        module = ast.parse(source)

        # Find the first class (should benefit from early exit)
        start = time.perf_counter()
        for _ in range(1000):
            _find_class_node(module, "TargetClass")
        elapsed_first = time.perf_counter() - start

        # Find the last class (no benefit from early exit since it's at the end)
        start = time.perf_counter()
        for _ in range(1000):
            cls_node = _find_class_node(module, f"OtherClass{num_classes - 1}")
        elapsed_last = time.perf_counter() - start

        assert cls_node is not None

        # First class lookup should be significantly faster due to early exit
        # (at least 2x faster since TargetClass is at position 0 of 500)
        assert elapsed_first < elapsed_last, (
            f"Early exit not effective: first={elapsed_first:.4f}s, last={elapsed_last:.4f}s"
        )



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
import pytest
from azure.mgmt.datamigration.aio import DataMigrationManagementClient

from devtools_testutils import AzureMgmtRecordedTestCase, RandomNameResourceGroupPreparer
from devtools_testutils.aio import recorded_by_proxy_async

AZURE_LOCATION = "eastus"


@pytest.mark.skip("you may need to update the auto-generated test case before run it")
class TestDataMigrationManagementDatabaseMigrationsSqlVmOperationsAsync(AzureMgmtRecordedTestCase):
    def setup_method(self, method):
        self.client = self.create_mgmt_client(DataMigrationManagementClient, is_async=True)

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_database_migrations_sql_vm_get(self, resource_group):
        response = await self.client.database_migrations_sql_vm.get(
            resource_group_name=resource_group.name,
            sql_virtual_machine_name="str",
            target_db_name="str",
            api_version="2025-06-30",
        )

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_database_migrations_sql_vm_begin_create_or_update(self, resource_group):
        response = await (
            await self.client.database_migrations_sql_vm.begin_create_or_update(
                resource_group_name=resource_group.name,
                sql_virtual_machine_name="str",
                target_db_name="str",
                parameters={
                    "id": "str",
                    "name": "str",
                    "properties": {
                        "kind": "SqlVm",
                        "backupConfiguration": {
                            "sourceLocation": {
                                "azureBlob": {
                                    "accountKey": "str",
                                    "authType": "str",
                                    "blobContainerName": "str",
                                    "identity": {
                                        "type": "str",
                                        "principalId": "str",
                                        "tenantId": "str",
                                        "userAssignedIdentities": {"str": {"clientId": "str", "principalId": "str"}},
                                    },
                                    "storageAccountResourceId": "str",
                                },
                                "fileShare": {"password": "str", "path": "str", "username": "str"},
                                "fileStorageType": "str",
                            },
                            "targetLocation": {"accountKey": "str", "storageAccountResourceId": "str"},
                        },
                        "endedOn": "2020-02-20 00:00:00",
                        "migrationFailureError": {"code": "str", "message": "str"},
                        "migrationOperationId": "str",
                        "migrationService": "str",
                        "migrationStatus": "str",
                        "migrationStatusDetails": {
                            "activeBackupSets": [
                                {
                                    "backupFinishDate": "2020-02-20 00:00:00",
                                    "backupSetId": "str",
                                    "backupStartDate": "2020-02-20 00:00:00",
                                    "backupType": "str",
                                    "familyCount": 0,
                                    "firstLSN": "str",
                                    "hasBackupChecksums": bool,
                                    "ignoreReasons": ["str"],
                                    "isBackupRestored": bool,
                                    "lastLSN": "str",
                                    "listOfBackupFiles": [
                                        {
                                            "copyDuration": 0,
                                            "copyThroughput": 0.0,
                                            "dataRead": 0,
                                            "dataWritten": 0,
                                            "familySequenceNumber": 0,
                                            "fileName": "str",
                                            "status": "str",
                                            "totalSize": 0,
                                        }
                                    ],
                                }
                            ],
                            "blobContainerName": "str",
                            "completeRestoreErrorMessage": "str",
                            "currentRestoringFilename": "str",
                            "fileUploadBlockingErrors": ["str"],
                            "fullBackupSetInfo": {
                                "backupFinishDate": "2020-02-20 00:00:00",
                                "backupSetId": "str",
                                "backupStartDate": "2020-02-20 00:00:00",
                                "backupType": "str",
                                "familyCount": 0,
                                "firstLSN": "str",
                                "hasBackupChecksums": bool,
                                "ignoreReasons": ["str"],
                                "isBackupRestored": bool,
                                "lastLSN": "str",
                                "listOfBackupFiles": [
                                    {
                                        "copyDuration": 0,
                                        "copyThroughput": 0.0,
                                        "dataRead": 0,
                                        "dataWritten": 0,
                                        "familySequenceNumber": 0,
                                        "fileName": "str",
                                        "status": "str",
                                        "totalSize": 0,
                                    }
                                ],
                            },
                            "invalidFiles": ["str"],
                            "isFullBackupRestored": bool,
                            "lastRestoredBackupSetInfo": {
                                "backupFinishDate": "2020-02-20 00:00:00",
                                "backupSetId": "str",
                                "backupStartDate": "2020-02-20 00:00:00",
                                "backupType": "str",
                                "familyCount": 0,
                                "firstLSN": "str",
                                "hasBackupChecksums": bool,
                                "ignoreReasons": ["str"],
                                "isBackupRestored": bool,
                                "lastLSN": "str",
                                "listOfBackupFiles": [
                                    {
                                        "copyDuration": 0,
                                        "copyThroughput": 0.0,
                                        "dataRead": 0,
                                        "dataWritten": 0,
                                        "familySequenceNumber": 0,
                                        "fileName": "str",
                                        "status": "str",
                                        "totalSize": 0,
                                    }
                                ],
                            },
                            "lastRestoredFilename": "str",
                            "migrationState": "str",
                            "pendingLogBackupsCount": 0,
                            "restoreBlockingReason": "str",
                        },
                        "offlineConfiguration": {"lastBackupName": "str", "offline": bool},
                        "provisioningError": "str",
                        "provisioningState": "str",
                        "scope": "str",
                        "sourceDatabaseName": "str",
                        "sourceServerName": "str",
                        "sourceSqlConnection": {
                            "authentication": "str",
                            "dataSource": "str",
                            "encryptConnection": bool,
                            "password": "str",
                            "trustServerCertificate": bool,
                            "userName": "str",
                        },
                        "startedOn": "2020-02-20 00:00:00",
                        "targetDatabaseCollation": "str",
                    },
                    "systemData": {
                        "createdAt": "2020-02-20 00:00:00",
                        "createdBy": "str",
                        "createdByType": "str",
                        "lastModifiedAt": "2020-02-20 00:00:00",
                        "lastModifiedBy": "str",
                        "lastModifiedByType": "str",
                    },
                    "type": "str",
                },
                api_version="2025-06-30",
            )
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_database_migrations_sql_vm_begin_delete(self, resource_group):
        response = await (
            await self.client.database_migrations_sql_vm.begin_delete(
                resource_group_name=resource_group.name,
                sql_virtual_machine_name="str",
                target_db_name="str",
                api_version="2025-06-30",
            )
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_database_migrations_sql_vm_begin_cancel(self, resource_group):
        response = await (
            await self.client.database_migrations_sql_vm.begin_cancel(
                resource_group_name=resource_group.name,
                sql_virtual_machine_name="str",
                target_db_name="str",
                parameters={"migrationOperationId": "str"},
                api_version="2025-06-30",
            )
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...

    @RandomNameResourceGroupPreparer(location=AZURE_LOCATION)
    @recorded_by_proxy_async
    async def test_database_migrations_sql_vm_begin_cutover(self, resource_group):
        response = await (
            await self.client.database_migrations_sql_vm.begin_cutover(
                resource_group_name=resource_group.name,
                sql_virtual_machine_name="str",
                target_db_name="str",
                parameters={"migrationOperationId": "str"},
                api_version="2025-06-30",
            )
        ).result()  # call '.result()' to poll until service return final result

        # please add some check logic here by yourself
        # ...



# pylint: disable=too-many-lines
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
# pylint: disable=useless-super-delegation

import datetime
from typing import Any, Mapping, Optional, TYPE_CHECKING, Union, overload

from .._utils.model_base import Model as _Model, rest_field

if TYPE_CHECKING:
    from .. import models as _models


class AcrAccessToken(_Model):
    """The ACR access token response containing the access token for authentication.

    :ivar access_token: The access token for performing authenticated requests.
    :vartype access_token: str
    """

    access_token: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The access token for performing authenticated requests."""

    @overload
    def __init__(
        self,
        *,
        access_token: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class AcrErrorInfo(_Model):
    """Error information.

    :ivar code: Error code.
    :vartype code: str
    :ivar message: Error message.
    :vartype message: str
    :ivar detail: Error details.
    :vartype detail: any
    """

    code: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Error code."""
    message: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Error message."""
    detail: Optional[Any] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Error details."""

    @overload
    def __init__(
        self,
        *,
        code: Optional[str] = None,
        message: Optional[str] = None,
        detail: Optional[Any] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class AcrErrors(_Model):
    """Acr error response describing why the operation failed.

    :ivar errors: Array of detailed error.
    :vartype errors: list[~azure.containerregistry._generated.models.AcrErrorInfo]
    """

    errors: Optional[list["_models.AcrErrorInfo"]] = rest_field(
        visibility=["read", "create", "update", "delete", "query"]
    )
    """Array of detailed error."""

    @overload
    def __init__(
        self,
        *,
        errors: Optional[list["_models.AcrErrorInfo"]] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class AcrManifests(_Model):
    """Manifest attributes.

    :ivar registry_login_server: Registry login server name. This is likely to be similar to
     {registry-name}.azurecr.io.
    :vartype registry_login_server: str
    :ivar repository: Image name.
    :vartype repository: str
    :ivar manifests: List of manifests.
    :vartype manifests: list[~azure.containerregistry._generated.models.ManifestAttributesBase]
    :ivar link: Link to the next page of results.
    :vartype link: str
    """

    registry_login_server: Optional[str] = rest_field(
        name="registry", visibility=["read", "create", "update", "delete", "query"]
    )
    """Registry login server name. This is likely to be similar to
     {registry-name}.azurecr.io."""
    repository: Optional[str] = rest_field(name="imageName", visibility=["read", "create", "update", "delete", "query"])
    """Image name."""
    manifests: Optional[list["_models.ManifestAttributesBase"]] = rest_field(
        visibility=["read", "create", "update", "delete", "query"]
    )
    """List of manifests."""
    link: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Link to the next page of results."""

    @overload
    def __init__(
        self,
        *,
        registry_login_server: Optional[str] = None,
        repository: Optional[str] = None,
        manifests: Optional[list["_models.ManifestAttributesBase"]] = None,
        link: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class AcrRefreshToken(_Model):
    """The ACR refresh token response containing the refresh token for authentication.

    :ivar refresh_token: The refresh token to be used for generating access tokens.
    :vartype refresh_token: str
    """

    refresh_token: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The refresh token to be used for generating access tokens."""

    @overload
    def __init__(
        self,
        *,
        refresh_token: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Annotations(_Model):
    """Additional information provided through arbitrary metadata.

    :ivar created: Date and time on which the image was built (string, date-time as defined by
     `https://tools.ietf.org/html/rfc3339#section-5.6
     <https://tools.ietf.org/html/rfc3339#section-5.6>`_).
    :vartype created: ~datetime.datetime
    :ivar authors: Contact details of the people or organization responsible for the image.
    :vartype authors: str
    :ivar url: URL to find more information on the image.
    :vartype url: str
    :ivar documentation: URL to get documentation on the image.
    :vartype documentation: str
    :ivar source: URL to get source code for building the image.
    :vartype source: str
    :ivar version: Version of the packaged software. The version MAY match a label or tag in the
     source code repository, may also be Semantic versioning-compatible.
    :vartype version: str
    :ivar revision: Source control revision identifier for the packaged software.
    :vartype revision: str
    :ivar vendor: Name of the distributing entity, organization or individual.
    :vartype vendor: str
    :ivar licenses: License(s) under which contained software is distributed as an SPDX License
     Expression.
    :vartype licenses: str
    :ivar name: Name of the reference for a target.
    :vartype name: str
    :ivar title: Human-readable title of the image.
    :vartype title: str
    :ivar description: Human-readable description of the software packaged in the image.
    :vartype description: str
    """

    created: Optional[datetime.datetime] = rest_field(
        name="org.opencontainers.image.created",
        visibility=["read", "create", "update", "delete", "query"],
        format="rfc3339",
    )
    """Date and time on which the image was built (string, date-time as defined by
     `https://tools.ietf.org/html/rfc3339#section-5.6
     <https://tools.ietf.org/html/rfc3339#section-5.6>`_)."""
    authors: Optional[str] = rest_field(
        name="org.opencontainers.image.authors",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Contact details of the people or organization responsible for the image."""
    url: Optional[str] = rest_field(
        name="org.opencontainers.image.url",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """URL to find more information on the image."""
    documentation: Optional[str] = rest_field(
        name="org.opencontainers.image.documentation",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """URL to get documentation on the image."""
    source: Optional[str] = rest_field(
        name="org.opencontainers.image.source",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """URL to get source code for building the image."""
    version: Optional[str] = rest_field(
        name="org.opencontainers.image.version",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Version of the packaged software. The version MAY match a label or tag in the
     source code repository, may also be Semantic versioning-compatible."""
    revision: Optional[str] = rest_field(
        name="org.opencontainers.image.revision",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Source control revision identifier for the packaged software."""
    vendor: Optional[str] = rest_field(
        name="org.opencontainers.image.vendor",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Name of the distributing entity, organization or individual."""
    licenses: Optional[str] = rest_field(
        name="org.opencontainers.image.licenses",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """License(s) under which contained software is distributed as an SPDX License
     Expression."""
    name: Optional[str] = rest_field(
        name="org.opencontainers.image.ref.name",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Name of the reference for a target."""
    title: Optional[str] = rest_field(
        name="org.opencontainers.image.title",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Human-readable title of the image."""
    description: Optional[str] = rest_field(
        name="org.opencontainers.image.description",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Human-readable description of the software packaged in the image."""

    @overload
    def __init__(
        self,
        *,
        created: Optional[datetime.datetime] = None,
        authors: Optional[str] = None,
        url: Optional[str] = None,
        documentation: Optional[str] = None,
        source: Optional[str] = None,
        version: Optional[str] = None,
        revision: Optional[str] = None,
        vendor: Optional[str] = None,
        licenses: Optional[str] = None,
        name: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class ArtifactManifestPlatform(_Model):
    """The artifact's platform, consisting of operating system and architecture.

    :ivar digest: Manifest digest. Required.
    :vartype digest: str
    :ivar architecture: CPU architecture. Known values are: "386", "amd64", "arm", "arm64", "mips",
     "mipsle", "mips64", "mips64le", "ppc64", "ppc64le", "riscv64", "s390x", and "wasm".
    :vartype architecture: str or ~azure.containerregistry._generated.models.ArtifactArchitecture
    :ivar operating_system: Operating system. Known values are: "aix", "android", "darwin",
     "dragonfly", "freebsd", "illumos", "ios", "js", "linux", "netbsd", "openbsd", "plan9",
     "solaris", and "windows".
    :vartype operating_system: str or
     ~azure.containerregistry._generated.models.ArtifactOperatingSystem
    """

    digest: str = rest_field(visibility=["read"])
    """Manifest digest. Required."""
    architecture: Optional[Union[str, "_models.ArtifactArchitecture"]] = rest_field(visibility=["read"])
    """CPU architecture. Known values are: \"386\", \"amd64\", \"arm\", \"arm64\", \"mips\",
     \"mipsle\", \"mips64\", \"mips64le\", \"ppc64\", \"ppc64le\", \"riscv64\", \"s390x\", and
     \"wasm\"."""
    operating_system: Optional[Union[str, "_models.ArtifactOperatingSystem"]] = rest_field(
        name="os", visibility=["read"]
    )
    """Operating system. Known values are: \"aix\", \"android\", \"darwin\", \"dragonfly\",
     \"freebsd\", \"illumos\", \"ios\", \"js\", \"linux\", \"netbsd\", \"openbsd\", \"plan9\",
     \"solaris\", and \"windows\"."""


class ArtifactManifestProperties(_Model):
    """Manifest attributes details.

    :ivar registry_login_server: Registry login server name. This is likely to be similar to
     {registry-name}.azurecr.io.
    :vartype registry_login_server: str
    :ivar repository_name: Repository name.
    :vartype repository_name: str
    :ivar manifest: Manifest attributes. Required.
    :vartype manifest: ~azure.containerregistry._generated.models.ManifestAttributesBase
    """

    registry_login_server: Optional[str] = rest_field(name="registry", visibility=["read"])
    """Registry login server name. This is likely to be similar to
     {registry-name}.azurecr.io."""
    repository_name: Optional[str] = rest_field(name="imageName", visibility=["read"])
    """Repository name."""
    manifest: "_models.ManifestAttributesBase" = rest_field(visibility=["read"])
    """Manifest attributes. Required."""


class ArtifactTagProperties(_Model):
    """Tag attributes.

    :ivar registry_login_server: Registry login server name. This is likely to be similar to
     {registry-name}.azurecr.io. Required.
    :vartype registry_login_server: str
    :ivar repository_name: Image name. Required.
    :vartype repository_name: str
    :ivar tag: List of tag attribute details. Required.
    :vartype tag: ~azure.containerregistry._generated.models.TagAttributesBase
    """

    registry_login_server: str = rest_field(name="registry", visibility=["read"])
    """Registry login server name. This is likely to be similar to
     {registry-name}.azurecr.io. Required."""
    repository_name: str = rest_field(name="imageName", visibility=["read"])
    """Image name. Required."""
    tag: "_models.TagAttributesBase" = rest_field(visibility=["read"])
    """List of tag attribute details. Required."""


class ContainerRepositoryProperties(_Model):
    """Properties of this repository.

    :ivar registry_login_server: Registry login server name. This is likely to be similar to
     {registry-name}.azurecr.io. Required.
    :vartype registry_login_server: str
    :ivar name: Image name. Required.
    :vartype name: str
    :ivar created_on: Image created time. Required.
    :vartype created_on: ~datetime.datetime
    :ivar last_updated_on: Image last update time. Required.
    :vartype last_updated_on: ~datetime.datetime
    :ivar manifest_count: Number of the manifests. Required.
    :vartype manifest_count: int
    :ivar tag_count: Number of the tags. Required.
    :vartype tag_count: int
    :ivar changeable_attributes: Writeable properties of the resource. Required.
    :vartype changeable_attributes:
     ~azure.containerregistry._generated.models.RepositoryWriteableProperties
    """

    registry_login_server: str = rest_field(name="registry", visibility=["read"])
    """Registry login server name. This is likely to be similar to
     {registry-name}.azurecr.io. Required."""
    name: str = rest_field(name="imageName", visibility=["read"])
    """Image name. Required."""
    created_on: datetime.datetime = rest_field(name="createdTime", visibility=["read"], format="rfc3339")
    """Image created time. Required."""
    last_updated_on: datetime.datetime = rest_field(name="lastUpdateTime", visibility=["read"], format="rfc3339")
    """Image last update time. Required."""
    manifest_count: int = rest_field(name="manifestCount", visibility=["read"])
    """Number of the manifests. Required."""
    tag_count: int = rest_field(name="tagCount", visibility=["read"])
    """Number of the tags. Required."""
    changeable_attributes: "_models.RepositoryWriteableProperties" = rest_field(
        name="changeableAttributes",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Writeable properties of the resource. Required."""

    @overload
    def __init__(
        self,
        *,
        changeable_attributes: "_models.RepositoryWriteableProperties",
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Descriptor(_Model):
    """Docker V2 image layer descriptor including config and layers.

    :ivar media_type: Layer media type.
    :vartype media_type: str
    :ivar size: Layer size.
    :vartype size: int
    :ivar digest: Layer digest.
    :vartype digest: str
    :ivar urls: Specifies a list of URIs from which this object may be downloaded.
    :vartype urls: list[str]
    :ivar annotations: Additional information provided through arbitrary metadata.
    :vartype annotations: ~azure.containerregistry._generated.models.Annotations
    """

    media_type: Optional[str] = rest_field(name="mediaType", visibility=["read", "create", "update", "delete", "query"])
    """Layer media type."""
    size: Optional[int] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Layer size."""
    digest: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Layer digest."""
    urls: Optional[list[str]] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Specifies a list of URIs from which this object may be downloaded."""
    annotations: Optional["_models.Annotations"] = rest_field(
        visibility=["read", "create", "update", "delete", "query"]
    )
    """Additional information provided through arbitrary metadata."""

    @overload
    def __init__(
        self,
        *,
        media_type: Optional[str] = None,
        size: Optional[int] = None,
        digest: Optional[str] = None,
        urls: Optional[list[str]] = None,
        annotations: Optional["_models.Annotations"] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class FsLayer(_Model):
    """Image layer information.

    :ivar blob_sum: SHA of an image layer.
    :vartype blob_sum: str
    """

    blob_sum: Optional[str] = rest_field(name="blobSum", visibility=["read", "create", "update", "delete", "query"])
    """SHA of an image layer."""

    @overload
    def __init__(
        self,
        *,
        blob_sum: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class History(_Model):
    """A list of unstructured historical data for v1 compatibility.

    :ivar v1_compatibility: The raw v1 compatibility information.
    :vartype v1_compatibility: str
    """

    v1_compatibility: Optional[str] = rest_field(
        name="v1Compatibility",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """The raw v1 compatibility information."""

    @overload
    def __init__(
        self,
        *,
        v1_compatibility: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class ImageSignature(_Model):
    """Signature of a signed manifest.

    :ivar header: A JSON web signature.
    :vartype header: ~azure.containerregistry._generated.models.JWK
    :ivar signature: A signature for the image manifest, signed by a libtrust private key.
    :vartype signature: str
    :ivar protected: The signed protected header.
    :vartype protected: str
    """

    header: Optional["_models.JWK"] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """A JSON web signature."""
    signature: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """A signature for the image manifest, signed by a libtrust private key."""
    protected: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The signed protected header."""

    @overload
    def __init__(
        self,
        *,
        header: Optional["_models.JWK"] = None,
        signature: Optional[str] = None,
        protected: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class JWK(_Model):
    """A JSON web signature.

    :ivar jwk: JSON web key parameter.
    :vartype jwk: ~azure.containerregistry._generated.models.JWKHeader
    :ivar alg: The algorithm used to sign or encrypt the JWT.
    :vartype alg: str
    """

    jwk: Optional["_models.JWKHeader"] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """JSON web key parameter."""
    alg: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The algorithm used to sign or encrypt the JWT."""

    @overload
    def __init__(
        self,
        *,
        jwk: Optional["_models.JWKHeader"] = None,
        alg: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class JWKHeader(_Model):
    """JSON web key parameter.

    :ivar crv: crv value.
    :vartype crv: str
    :ivar kid: kid value.
    :vartype kid: str
    :ivar kty: kty value.
    :vartype kty: str
    :ivar x: x value.
    :vartype x: str
    :ivar y: y value.
    :vartype y: str
    """

    crv: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """crv value."""
    kid: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """kid value."""
    kty: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """kty value."""
    x: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """x value."""
    y: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """y value."""

    @overload
    def __init__(
        self,
        *,
        crv: Optional[str] = None,
        kid: Optional[str] = None,
        kty: Optional[str] = None,
        x: Optional[str] = None,
        y: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Manifest(_Model):
    """Returns the requested manifest file.

    :ivar schema_version: Schema version.
    :vartype schema_version: int
    """

    schema_version: Optional[int] = rest_field(
        name="schemaVersion", visibility=["read", "create", "update", "delete", "query"]
    )
    """Schema version."""

    @overload
    def __init__(
        self,
        *,
        schema_version: Optional[int] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class ManifestAttributesBase(_Model):
    """Manifest details.

    :ivar digest: Manifest. Required.
    :vartype digest: str
    :ivar size: Image size.
    :vartype size: int
    :ivar created_on: Created time. Required.
    :vartype created_on: ~datetime.datetime
    :ivar last_updated_on: Last update time. Required.
    :vartype last_updated_on: ~datetime.datetime
    :ivar architecture: CPU architecture. Known values are: "386", "amd64", "arm", "arm64", "mips",
     "mipsle", "mips64", "mips64le", "ppc64", "ppc64le", "riscv64", "s390x", and "wasm".
    :vartype architecture: str or ~azure.containerregistry._generated.models.ArtifactArchitecture
    :ivar operating_system: Operating system. Known values are: "aix", "android", "darwin",
     "dragonfly", "freebsd", "illumos", "ios", "js", "linux", "netbsd", "openbsd", "plan9",
     "solaris", and "windows".
    :vartype operating_system: str or
     ~azure.containerregistry._generated.models.ArtifactOperatingSystem
    :ivar related_artifacts: List of artifacts that are referenced by this manifest list, with
     information
     about the platform each supports.  This list will be empty if this is a leaf
     manifest and not a manifest list.
    :vartype related_artifacts:
     list[~azure.containerregistry._generated.models.ArtifactManifestPlatform]
    :ivar config_media_type: Config blob media type.
    :vartype config_media_type: str
    :ivar tags: List of tags.
    :vartype tags: list[str]
    :ivar changeable_attributes: Writeable properties of the resource.
    :vartype changeable_attributes:
     ~azure.containerregistry._generated.models.ManifestWriteableProperties
    """

    digest: str = rest_field(visibility=["read"])
    """Manifest. Required."""
    size: Optional[int] = rest_field(name="imageSize", visibility=["read"])
    """Image size."""
    created_on: datetime.datetime = rest_field(name="createdTime", visibility=["read"], format="rfc3339")
    """Created time. Required."""
    last_updated_on: datetime.datetime = rest_field(name="lastUpdateTime", visibility=["read"], format="rfc3339")
    """Last update time. Required."""
    architecture: Optional[Union[str, "_models.ArtifactArchitecture"]] = rest_field(visibility=["read"])
    """CPU architecture. Known values are: \"386\", \"amd64\", \"arm\", \"arm64\", \"mips\",
     \"mipsle\", \"mips64\", \"mips64le\", \"ppc64\", \"ppc64le\", \"riscv64\", \"s390x\", and
     \"wasm\"."""
    operating_system: Optional[Union[str, "_models.ArtifactOperatingSystem"]] = rest_field(
        name="os", visibility=["read"]
    )
    """Operating system. Known values are: \"aix\", \"android\", \"darwin\", \"dragonfly\",
     \"freebsd\", \"illumos\", \"ios\", \"js\", \"linux\", \"netbsd\", \"openbsd\", \"plan9\",
     \"solaris\", and \"windows\"."""
    related_artifacts: Optional[list["_models.ArtifactManifestPlatform"]] = rest_field(
        name="references", visibility=["read"]
    )
    """List of artifacts that are referenced by this manifest list, with information
     about the platform each supports.  This list will be empty if this is a leaf
     manifest and not a manifest list."""
    config_media_type: Optional[str] = rest_field(
        name="configMediaType",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Config blob media type."""
    tags: Optional[list[str]] = rest_field(visibility=["read"])
    """List of tags."""
    changeable_attributes: Optional["_models.ManifestWriteableProperties"] = rest_field(
        name="changeableAttributes",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Writeable properties of the resource."""

    @overload
    def __init__(
        self,
        *,
        config_media_type: Optional[str] = None,
        changeable_attributes: Optional["_models.ManifestWriteableProperties"] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class ManifestListAttributes(_Model):
    """Attributes of a manifest in a manifest list.

    :ivar media_type: The MIME type of the referenced object. This will generally be
     application/vnd.docker.image.manifest.v2+json, but it could also be
     application/vnd.docker.image.manifest.v1+json.
    :vartype media_type: str
    :ivar size: The size in bytes of the object.
    :vartype size: int
    :ivar digest: The digest of the content, as defined by the Registry V2 HTTP API Specification.
    :vartype digest: str
    :ivar platform: The platform object describes the platform which the image in the manifest runs
     on. A full list of valid operating system and architecture values are listed in
     the Go language documentation for $GOOS and $GOARCH.
    :vartype platform: ~azure.containerregistry._generated.models.Platform
    """

    media_type: Optional[str] = rest_field(name="mediaType", visibility=["read", "create", "update", "delete", "query"])
    """The MIME type of the referenced object. This will generally be
     application/vnd.docker.image.manifest.v2+json, but it could also be
     application/vnd.docker.image.manifest.v1+json."""
    size: Optional[int] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The size in bytes of the object."""
    digest: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The digest of the content, as defined by the Registry V2 HTTP API Specification."""
    platform: Optional["_models.Platform"] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The platform object describes the platform which the image in the manifest runs
     on. A full list of valid operating system and architecture values are listed in
     the Go language documentation for $GOOS and $GOARCH."""

    @overload
    def __init__(
        self,
        *,
        media_type: Optional[str] = None,
        size: Optional[int] = None,
        digest: Optional[str] = None,
        platform: Optional["_models.Platform"] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class ManifestWrapper(Manifest):
    """Returns the requested manifest file.

    :ivar schema_version: Schema version.
    :vartype schema_version: int
    :ivar media_type: Media type for this Manifest.
    :vartype media_type: str
    :ivar manifests: (ManifestList, OCIIndex) List of V2 image layer information.
    :vartype manifests: list[~azure.containerregistry._generated.models.ManifestListAttributes]
    :ivar config: (V2, OCI) Image config descriptor.
    :vartype config: ~azure.containerregistry._generated.models.Descriptor
    :ivar layers: (V2, OCI) List of V2 image layer information.
    :vartype layers: list[~azure.containerregistry._generated.models.Descriptor]
    :ivar annotations: (OCI, OCIIndex) Additional metadata.
    :vartype annotations: ~azure.containerregistry._generated.models.Annotations
    :ivar architecture: (V1) CPU architecture.
    :vartype architecture: str
    :ivar name: (V1) Image name.
    :vartype name: str
    :ivar tag: (V1) Image tag.
    :vartype tag: str
    :ivar fs_layers: (V1) List of layer information.
    :vartype fs_layers: list[~azure.containerregistry._generated.models.FsLayer]
    :ivar history: (V1) Image history.
    :vartype history: list[~azure.containerregistry._generated.models.History]
    :ivar signatures: (V1) Image signature.
    :vartype signatures: list[~azure.containerregistry._generated.models.ImageSignature]
    """

    media_type: Optional[str] = rest_field(name="mediaType", visibility=["read", "create", "update", "delete", "query"])
    """Media type for this Manifest."""
    manifests: Optional[list["_models.ManifestListAttributes"]] = rest_field(
        visibility=["read", "create", "update", "delete", "query"]
    )
    """(ManifestList, OCIIndex) List of V2 image layer information."""
    config: Optional["_models.Descriptor"] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """(V2, OCI) Image config descriptor."""
    layers: Optional[list["_models.Descriptor"]] = rest_field(
        visibility=["read", "create", "update", "delete", "query"]
    )
    """(V2, OCI) List of V2 image layer information."""
    annotations: Optional["_models.Annotations"] = rest_field(
        visibility=["read", "create", "update", "delete", "query"]
    )
    """(OCI, OCIIndex) Additional metadata."""
    architecture: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """(V1) CPU architecture."""
    name: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """(V1) Image name."""
    tag: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """(V1) Image tag."""
    fs_layers: Optional[list["_models.FsLayer"]] = rest_field(
        name="fsLayers", visibility=["read", "create", "update", "delete", "query"]
    )
    """(V1) List of layer information."""
    history: Optional[list["_models.History"]] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """(V1) Image history."""
    signatures: Optional[list["_models.ImageSignature"]] = rest_field(
        visibility=["read", "create", "update", "delete", "query"]
    )
    """(V1) Image signature."""

    @overload
    def __init__(
        self,
        *,
        schema_version: Optional[int] = None,
        media_type: Optional[str] = None,
        manifests: Optional[list["_models.ManifestListAttributes"]] = None,
        config: Optional["_models.Descriptor"] = None,
        layers: Optional[list["_models.Descriptor"]] = None,
        annotations: Optional["_models.Annotations"] = None,
        architecture: Optional[str] = None,
        name: Optional[str] = None,
        tag: Optional[str] = None,
        fs_layers: Optional[list["_models.FsLayer"]] = None,
        history: Optional[list["_models.History"]] = None,
        signatures: Optional[list["_models.ImageSignature"]] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class ManifestWriteableProperties(_Model):
    """Changeable attributes.

    :ivar can_delete: Delete enabled.
    :vartype can_delete: bool
    :ivar can_write: Write enabled.
    :vartype can_write: bool
    :ivar can_list: List enabled.
    :vartype can_list: bool
    :ivar can_read: Read enabled.
    :vartype can_read: bool
    """

    can_delete: Optional[bool] = rest_field(
        name="deleteEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """Delete enabled."""
    can_write: Optional[bool] = rest_field(
        name="writeEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """Write enabled."""
    can_list: Optional[bool] = rest_field(
        name="listEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """List enabled."""
    can_read: Optional[bool] = rest_field(
        name="readEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """Read enabled."""

    @overload
    def __init__(
        self,
        *,
        can_delete: Optional[bool] = None,
        can_write: Optional[bool] = None,
        can_list: Optional[bool] = None,
        can_read: Optional[bool] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class MultipartBodyParameter(_Model):
    """The multipart body parameter for AAD token exchange.

    :ivar grant_type: Can take a value of access_token_refresh_token, or access_token, or
     refresh_token. Required. Known values are: "access_token_refresh_token", "access_token", and
     "refresh_token".
    :vartype grant_type: str or
     ~azure.containerregistry._generated.models.PostContentSchemaGrantType
    :ivar service: Indicates the name of your Azure container registry. Required.
    :vartype service: str
    :ivar tenant: AAD tenant associated to the AAD credentials.
    :vartype tenant: str
    :ivar refresh_token: AAD refresh token, mandatory when grant_type is access_token_refresh_token
     or
     refresh_token.
    :vartype refresh_token: str
    :ivar access_token: AAD access token, mandatory when grant_type is access_token_refresh_token
     or
     access_token.
    :vartype access_token: str
    """

    grant_type: Union[str, "_models.PostContentSchemaGrantType"] = rest_field(
        name="grantType", visibility=["read", "create", "update", "delete", "query"]
    )
    """Can take a value of access_token_refresh_token, or access_token, or
     refresh_token. Required. Known values are: \"access_token_refresh_token\", \"access_token\",
     and \"refresh_token\"."""
    service: str = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Indicates the name of your Azure container registry. Required."""
    tenant: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """AAD tenant associated to the AAD credentials."""
    refresh_token: Optional[str] = rest_field(
        name="refreshToken", visibility=["read", "create", "update", "delete", "query"]
    )
    """AAD refresh token, mandatory when grant_type is access_token_refresh_token or
     refresh_token."""
    access_token: Optional[str] = rest_field(
        name="accessToken", visibility=["read", "create", "update", "delete", "query"]
    )
    """AAD access token, mandatory when grant_type is access_token_refresh_token or
     access_token."""

    @overload
    def __init__(
        self,
        *,
        grant_type: Union[str, "_models.PostContentSchemaGrantType"],
        service: str,
        tenant: Optional[str] = None,
        refresh_token: Optional[str] = None,
        access_token: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Platform(_Model):
    """The platform object describes the platform which the image in the manifest runs
    on. A full list of valid operating system and architecture values are listed in
    the Go language documentation for $GOOS and $GOARCH.

    :ivar architecture: Specifies the CPU architecture, for example amd64 or ppc64le.
    :vartype architecture: str
    :ivar os: The os field specifies the operating system, for example linux or windows.
    :vartype os: str
    :ivar os_version: The optional os.version field specifies the operating system version, for
     example 10.0.10586.
    :vartype os_version: str
    :ivar os_features: The optional os.features field specifies an array of strings, each listing a
     required OS feature (for example on Windows win32k.
    :vartype os_features: list[str]
    :ivar variant: The optional variant field specifies a variant of the CPU, for example armv6l
     to specify a particular CPU variant of the ARM CPU.
    :vartype variant: str
    :ivar features: The optional features field specifies an array of strings, each listing a
     required CPU feature (for example sse4 or aes.
    :vartype features: list[str]
    """

    architecture: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Specifies the CPU architecture, for example amd64 or ppc64le."""
    os: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The os field specifies the operating system, for example linux or windows."""
    os_version: Optional[str] = rest_field(
        name="os.version", visibility=["read", "create", "update", "delete", "query"]
    )
    """The optional os.version field specifies the operating system version, for
     example 10.0.10586."""
    os_features: Optional[list[str]] = rest_field(
        name="os.features", visibility=["read", "create", "update", "delete", "query"]
    )
    """The optional os.features field specifies an array of strings, each listing a
     required OS feature (for example on Windows win32k."""
    variant: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The optional variant field specifies a variant of the CPU, for example armv6l
     to specify a particular CPU variant of the ARM CPU."""
    features: Optional[list[str]] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The optional features field specifies an array of strings, each listing a
     required CPU feature (for example sse4 or aes."""

    @overload
    def __init__(
        self,
        *,
        architecture: Optional[str] = None,
        os: Optional[str] = None,
        os_version: Optional[str] = None,
        os_features: Optional[list[str]] = None,
        variant: Optional[str] = None,
        features: Optional[list[str]] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class RepositoryWriteableProperties(_Model):
    """Changeable attributes for Repository.

    :ivar can_delete: Delete enabled.
    :vartype can_delete: bool
    :ivar can_write: Write enabled.
    :vartype can_write: bool
    :ivar can_list: List enabled.
    :vartype can_list: bool
    :ivar can_read: Read enabled.
    :vartype can_read: bool
    """

    can_delete: Optional[bool] = rest_field(
        name="deleteEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """Delete enabled."""
    can_write: Optional[bool] = rest_field(
        name="writeEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """Write enabled."""
    can_list: Optional[bool] = rest_field(
        name="listEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """List enabled."""
    can_read: Optional[bool] = rest_field(
        name="readEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """Read enabled."""

    @overload
    def __init__(
        self,
        *,
        can_delete: Optional[bool] = None,
        can_write: Optional[bool] = None,
        can_list: Optional[bool] = None,
        can_read: Optional[bool] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class TagAttributesBase(_Model):
    """Tag attribute details.

    :ivar name: Tag name. Required.
    :vartype name: str
    :ivar digest: Tag digest. Required.
    :vartype digest: str
    :ivar created_on: Tag created time. Required.
    :vartype created_on: ~datetime.datetime
    :ivar last_updated_on: Tag last update time. Required.
    :vartype last_updated_on: ~datetime.datetime
    :ivar signed: Is signed.
    :vartype signed: bool
    :ivar changeable_attributes: Writeable properties of the resource. Required.
    :vartype changeable_attributes:
     ~azure.containerregistry._generated.models.TagWriteableProperties
    """

    name: str = rest_field(visibility=["read"])
    """Tag name. Required."""
    digest: str = rest_field(visibility=["read"])
    """Tag digest. Required."""
    created_on: datetime.datetime = rest_field(name="createdTime", visibility=["read"], format="rfc3339")
    """Tag created time. Required."""
    last_updated_on: datetime.datetime = rest_field(name="lastUpdateTime", visibility=["read"], format="rfc3339")
    """Tag last update time. Required."""
    signed: Optional[bool] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Is signed."""
    changeable_attributes: "_models.TagWriteableProperties" = rest_field(
        name="changeableAttributes",
        visibility=["read", "create", "update", "delete", "query"],
    )
    """Writeable properties of the resource. Required."""

    @overload
    def __init__(
        self,
        *,
        changeable_attributes: "_models.TagWriteableProperties",
        signed: Optional[bool] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class TagList(_Model):
    """List of tag details.

    :ivar registry_login_server: Registry login server name. This is likely to be similar to
     {registry-name}.azurecr.io. Required.
    :vartype registry_login_server: str
    :ivar repository: Image name. Required.
    :vartype repository: str
    :ivar tag_attribute_bases: List of tag attribute details. Required.
    :vartype tag_attribute_bases:
     list[~azure.containerregistry._generated.models.TagAttributesBase]
    :ivar link: Link to the next page of results.
    :vartype link: str
    """

    registry_login_server: str = rest_field(name="registry", visibility=["read", "create", "update", "delete", "query"])
    """Registry login server name. This is likely to be similar to
     {registry-name}.azurecr.io. Required."""
    repository: str = rest_field(name="imageName", visibility=["read", "create", "update", "delete", "query"])
    """Image name. Required."""
    tag_attribute_bases: list["_models.TagAttributesBase"] = rest_field(
        name="tags", visibility=["read", "create", "update", "delete", "query"]
    )
    """List of tag attribute details. Required."""
    link: Optional[str] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """Link to the next page of results."""

    @overload
    def __init__(
        self,
        *,
        registry_login_server: str,
        repository: str,
        tag_attribute_bases: list["_models.TagAttributesBase"],
        link: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class TagWriteableProperties(_Model):
    """Changeable attributes.

    :ivar can_delete: Delete enabled.
    :vartype can_delete: bool
    :ivar can_write: Write enabled.
    :vartype can_write: bool
    :ivar can_list: List enabled.
    :vartype can_list: bool
    :ivar can_read: Read enabled.
    :vartype can_read: bool
    """

    can_delete: Optional[bool] = rest_field(
        name="deleteEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """Delete enabled."""
    can_write: Optional[bool] = rest_field(
        name="writeEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """Write enabled."""
    can_list: Optional[bool] = rest_field(
        name="listEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """List enabled."""
    can_read: Optional[bool] = rest_field(
        name="readEnabled", visibility=["read", "create", "update", "delete", "query"]
    )
    """Read enabled."""

    @overload
    def __init__(
        self,
        *,
        can_delete: Optional[bool] = None,
        can_write: Optional[bool] = None,
        can_list: Optional[bool] = None,
        can_read: Optional[bool] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)



# pylint: disable=line-too-long,useless-suppression
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from copy import deepcopy
from typing import Any, Awaitable, TYPE_CHECKING, Union
from typing_extensions import Self

from azure.core import AsyncPipelineClient
from azure.core.credentials import AzureKeyCredential
from azure.core.pipeline import policies
from azure.core.rest import AsyncHttpResponse, HttpRequest

from .._utils.serialization import Deserializer, Serializer
from ._configuration import TextAnalysisClientConfiguration
from ._operations import _TextAnalysisClientOperationsMixin

if TYPE_CHECKING:
    from azure.core.credentials_async import AsyncTokenCredential


class TextAnalysisClient(_TextAnalysisClientOperationsMixin):
    """The language service API is a suite of natural language processing (NLP) skills built with
    best-in-class Microsoft machine learning algorithms.  The API can be used to analyze
    unstructured text for tasks such as sentiment analysis, key phrase extraction, language
    detection and question answering. Further documentation can be found in <a
    href=\\"`https://learn.microsoft.com/azure/cognitive-services/language-service/overview\\">https://learn.microsoft.com/azure/cognitive-services/language-service/overview
    <https://learn.microsoft.com/azure/cognitive-services/language-service/overview\\">https://learn.microsoft.com/azure/cognitive-services/language-service/overview>`_</a>.0.

    :param endpoint: Supported Cognitive Services endpoint (e.g.,
     https://<resource-name>.api.cognitiveservices.azure.com). Required.
    :type endpoint: str
    :param credential: Credential used to authenticate requests to the service. Is either a key
     credential type or a token credential type. Required.
    :type credential: ~azure.core.credentials.AzureKeyCredential or
     ~azure.core.credentials_async.AsyncTokenCredential
    :keyword api_version: The API version to use for this operation. Default value is
     "2025-11-15-preview". Note that overriding this default value may result in unsupported
     behavior.
    :paramtype api_version: str
    :keyword int polling_interval: Default waiting time between two polls for LRO operations if no
     Retry-After header is present.
    """

    def __init__(
        self, endpoint: str, credential: Union[AzureKeyCredential, "AsyncTokenCredential"], **kwargs: Any
    ) -> None:
        _endpoint = "{Endpoint}/language"
        self._config = TextAnalysisClientConfiguration(endpoint=endpoint, credential=credential, **kwargs)

        _policies = kwargs.pop("policies", None)
        if _policies is None:
            _policies = [
                policies.RequestIdPolicy(**kwargs),
                self._config.headers_policy,
                self._config.user_agent_policy,
                self._config.proxy_policy,
                policies.ContentDecodePolicy(**kwargs),
                self._config.redirect_policy,
                self._config.retry_policy,
                self._config.authentication_policy,
                self._config.custom_hook_policy,
                self._config.logging_policy,
                policies.DistributedTracingPolicy(**kwargs),
                policies.SensitiveHeaderCleanupPolicy(**kwargs) if self._config.redirect_policy else None,
                self._config.http_logging_policy,
            ]
        self._client: AsyncPipelineClient = AsyncPipelineClient(base_url=_endpoint, policies=_policies, **kwargs)

        self._serialize = Serializer()
        self._deserialize = Deserializer()
        self._serialize.client_side_validation = False

    def send_request(
        self, request: HttpRequest, *, stream: bool = False, **kwargs: Any
    ) -> Awaitable[AsyncHttpResponse]:
        """Runs the network request through the client's chained policies.

        >>> from azure.core.rest import HttpRequest
        >>> request = HttpRequest("GET", "https://www.example.org/")
        <HttpRequest [GET], url: 'https://www.example.org/'>
        >>> response = await client.send_request(request)
        <AsyncHttpResponse: 200 OK>

        For more information on this code flow, see https://aka.ms/azsdk/dpcodegen/python/send_request

        :param request: The network request you want to make. Required.
        :type request: ~azure.core.rest.HttpRequest
        :keyword bool stream: Whether the response payload will be streamed. Defaults to False.
        :return: The response of your network call. Does not do error handling on your response.
        :rtype: ~azure.core.rest.AsyncHttpResponse
        """

        request_copy = deepcopy(request)
        path_format_arguments = {
            "Endpoint": self._serialize.url("self._config.endpoint", self._config.endpoint, "str", skip_quote=True),
        }

        request_copy.url = self._client.format_url(request_copy.url, **path_format_arguments)
        return self._client.send_request(request_copy, stream=stream, **kwargs)  # type: ignore

    async def close(self) -> None:
        await self._client.close()

    async def __aenter__(self) -> Self:
        await self._client.__aenter__()
        return self

    async def __aexit__(self, *exc_details: Any) -> None:
        await self._client.__aexit__(*exc_details)



import argparse
import asyncio
import glob
import os
import sys
import time
import signal
import shutil
import subprocess
import re
from dataclasses import dataclass
from typing import IO, List, Optional

from ci_tools.functions import discover_targeted_packages
from ci_tools.variables import in_ci
from ci_tools.scenario.generation import build_whl_for_req, replace_dev_reqs
from ci_tools.logging import configure_logging, logger
from ci_tools.environment_exclusions import is_check_enabled, CHECK_DEFAULTS
from ci_tools.parsing import get_config_setting
from devtools_testutils.proxy_startup import prepare_local_tool
from packaging.requirements import Requirement

root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
ISOLATE_DIRS_TO_CLEAN: List[str] = []


@dataclass
class CheckResult:
    package: str
    check: str
    exit_code: int
    duration: float
    stdout: str
    stderr: str


@dataclass
class ProxyProcess:
    port: int
    process: subprocess.Popen
    log_handle: Optional[IO[str]]


PROXY_STATUS_SUFFIX = "/Info/Available"
PROXY_STARTUP_TIMEOUT = 60
BASE_PROXY_PORT = 5050
# Checks implemented via InstallAndTest all require shared recording restore behavior.
INSTALL_AND_TEST_CHECKS = {
    "whl",
    "whl_no_aio",
    "sdist",
    "devtest",
    "optional",
    "import_all",
    "latestdependency",
    "mindependency",
}
SHARED_RESTORE_ENV = "__shared_restore__"


def _cleanup_isolate_dirs() -> None:
    if not ISOLATE_DIRS_TO_CLEAN:
        return

    for path in ISOLATE_DIRS_TO_CLEAN:
        if not path:
            continue
        if os.path.exists(path):
            try:
                shutil.rmtree(path)
            except Exception:
                logger.warning(f"Failed to remove isolate dir {path}")
    ISOLATE_DIRS_TO_CLEAN.clear()


def _normalize_newlines(text: str) -> str:
    return text.replace("\r\n", "\n").replace("\r", "\n")


def _checks_require_recording_restore(checks: List[str]) -> bool:
    return any(check in INSTALL_AND_TEST_CHECKS for check in checks)


def _compare_req_to_injected_reqs(parsed_req, injected_packages: List[str]) -> bool:
    if parsed_req is None:
        return False
    return any(parsed_req.name in req for req in injected_packages)


def _inject_custom_reqs(req_file: str, injected_packages: str, package_dir: str) -> None:
    req_lines = []
    injected_list = [p for p in re.split(r"[\s,]", injected_packages) if p]

    if not injected_list:
        return

    # Entries prefixed with '!' are exclusion-only: they remove matching packages
    # from dev_requirements but are not themselves installed.
    excluded = [p[1:] for p in injected_list if p.startswith("!")]
    installable = [p for p in injected_list if not p.startswith("!")]
    # Build a combined list for filtering (both injected installs and exclusions)
    all_filter_names = installable + excluded

    logger.info(f"Adding custom packages to requirements for {package_dir}")
    with open(req_file, "r") as handle:
        for line in handle:
            logger.info(f"Attempting to parse {line}")
            try:
                parsed_req = Requirement(line.strip())
            except Exception as exc:
                logger.error(exc)
                parsed_req = None
            req_lines.append((line, parsed_req))

    if req_lines:
        all_adjustments = installable + [
            line_tuple[0].strip()
            for line_tuple in req_lines
            if line_tuple[0].strip() and not _compare_req_to_injected_reqs(line_tuple[1], all_filter_names)
        ]
    else:
        all_adjustments = installable

    logger.info(f"Generated Custom Reqs: {req_lines}")

    with open(req_file, "w") as handle:
        handle.write("\n".join(all_adjustments))


async def run_check(
    semaphore: asyncio.Semaphore,
    package: str,
    check: str,
    base_args: List[str],
    idx: int,
    total: int,
    proxy_port: int,
    mark_arg: Optional[str],
    dest_dir: Optional[str] = None,
    service: Optional[str] = None,
    python_version: Optional[str] = None,
) -> CheckResult:
    """Run a single check (subprocess) within a concurrency semaphore, capturing output and timing.

    :param semaphore: Concurrency limiter used to bound concurrent checks.
    :type semaphore: asyncio.Semaphore
    :param package: Absolute path to the package directory used as the subprocess cwd.
    :type package: str
    :param check: The check (subcommand) name for the azpysdk CLI to invoke.
    :type check: str
    :param base_args: Common argument list prefix (e.g. ``[sys.executable, "-m", "azpysdk.main"]``).
    :type base_args: List[str]
    :param idx: Sequence number for logging (1-based index of this task).
    :type idx: int
    :param total: Total number of tasks (used for logging progress).
    :type total: int
    :param proxy_port: Dedicated proxy port assigned to this check instance.
    :type proxy_port: int
    :returns: A :class:`CheckResult` describing exit code, duration and captured output.
    :rtype: CheckResult
    """
    async with semaphore:
        start = time.time()
        cmd = base_args + [check, "--isolate", package]
        if python_version:
            cmd += ["--python", python_version]
        if service:
            cmd += ["--service", service]
        if mark_arg:
            cmd += ["--mark_arg", mark_arg]
        if dest_dir and check == "apistub":
            cmd += ["--dest-dir", dest_dir]
        logger.info(f"[START {idx}/{total}] {check} :: {package}\nCMD: {' '.join(cmd)}")
        env = os.environ.copy()
        env["PROXY_URL"] = f"http://localhost:{proxy_port}"

        if in_ci():
            env["PROXY_ASSETS_FOLDER"] = os.path.join(root_dir, ".assets_distributed", str(proxy_port))
        try:
            logger.info(" ".join(cmd))
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                cwd=package,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
                env=env,
            )
        except Exception as ex:  # subprocess failed to launch
            logger.error(f"Failed to start check {check} for {package}: {ex}")
            return CheckResult(package, check, 127, 0.0, "", str(ex))

        stdout_b, stderr_b = await proc.communicate()
        duration = time.time() - start
        stdout = stdout_b.decode(errors="replace")
        stderr = stderr_b.decode(errors="replace")
        exit_code = proc.returncode or 0
        status = "OK" if exit_code == 0 else f"FAIL({exit_code})"
        logger.info(f"[END   {idx}/{total}] {check} :: {package} -> {status} in {duration:.2f}s")
        # Print captured output after completion to avoid interleaving
        header = f"===== OUTPUT: {check} :: {package} (exit {exit_code}) ====="
        trailer = "=" * len(header)
        if in_ci():
            print(f"##[group]{package} :: {check} :: {exit_code}")

        if stdout:
            print(header)
            print(_normalize_newlines(stdout).rstrip())
            print(trailer)
        if stderr:
            print(header.replace("OUTPUT", "STDERR"))
            print(_normalize_newlines(stderr).rstrip())
            print(trailer)

        if in_ci():
            print("##[endgroup]")

        # if we have any output collections to complete, do so now here

        # finally, we need to clean up any temp dirs created by --isolate
        if in_ci():
            package_name = os.path.basename(os.path.normpath(package))
            venv_pkg_root = os.path.join(root_dir, ".venv", package_name)
            # match both .venv_{check} and version-qualified .venv_{check}_py311 etc.
            for d in glob.glob(os.path.join(venv_pkg_root, f".venv_{check}*")):
                ISOLATE_DIRS_TO_CLEAN.append(d)
        return CheckResult(package, check, exit_code, duration, stdout, stderr)


def summarize(results: List[CheckResult]) -> int:
    """Print a compact summary table and return the worst exit code.

    The function prints a human-readable table to stdout showing package, check, status and
    duration. It returns the highest (worst) exit code from the provided results.

    :param results: List of :class:`CheckResult` objects to summarize.
    :type results: List[CheckResult]
    :returns: The maximum exit code found in ``results`` (0 if all passed).
    :rtype: int
    """
    # Compute column widths
    pkg_w = max((len(r.package) for r in results), default=7)
    chk_w = max((len(r.check) for r in results), default=5)
    header = f"{'PACKAGE'.ljust(pkg_w)}  {'CHECK'.ljust(chk_w)}  STATUS  DURATION(s)"
    print("\n=== SUMMARY ===")
    print(header)
    print("-" * len(header))
    for r in sorted(results, key=lambda x: (x.exit_code != 0, x.package, x.check)):
        status = "OK" if r.exit_code == 0 else f"FAIL({r.exit_code})"
        print(f"{r.package.ljust(pkg_w)}  {r.check.ljust(chk_w)}  {status.ljust(8)}  {r.duration:>10.2f}")
    worst = max((r.exit_code for r in results), default=0)
    failed = [r for r in results if r.exit_code != 0]
    print(f"\nTotal checks: {len(results)} | Failed: {len(failed)} | Worst exit code: {worst}")
    return worst


async def run_all_checks(
    packages,
    checks,
    max_parallel,
    wheel_dir,
    mark_arg: Optional[str],
    injected_packages: str,
    dest_dir: Optional[str] = None,
    service: Optional[str] = None,
):
    """Run all checks for all packages concurrently and return the worst exit code.

    :param packages: Iterable of package paths to run checks against.
    :type packages: Iterable[str]
    :param checks: List of check names to execute for each package.
    :type checks: List[str]
    :param max_parallel: Maximum number of concurrent checks to run.
    :type max_parallel: int
    :param wheel_dir: The directory where wheels should be located and stored when built.
        In CI should correspond to `$(Build.ArtifactStagingDirectory)`.
    :type wheel_dir: str
    :returns: The worst exit code from all checks (0 if all passed).
    :rtype: int
    """
    base_args = [sys.executable, "-m", "azpysdk.main"]
    tasks = []
    semaphore = asyncio.Semaphore(max_parallel)
    combos = [(p, c) for p in packages for c in checks]
    scheduled: List[tuple] = []

    test_tools_path = os.path.join(root_dir, "eng", "test_tools.txt")
    dependency_tools_path = os.path.join(root_dir, "eng", "dependency_tools.txt")

    if in_ci():
        logger.info("Replacing relative requirements in eng/test_tools.txt with prebuilt wheels.")
        replace_dev_reqs(test_tools_path, root_dir, wheel_dir)

        logger.info("Replacing relative requirements in eng/dependency_tools.txt with prebuilt wheels.")
        replace_dev_reqs(dependency_tools_path, root_dir, wheel_dir)

    for pkg in packages:
        destination_dev_req = os.path.join(pkg, "dev_requirements.txt")

        logger.info(f"Replacing dev requirements w/ path {destination_dev_req}")
        if not os.path.exists(destination_dev_req):
            logger.info("No dev_requirements present.")
            with open(destination_dev_req, "w+") as file:
                file.write("\n")

        if in_ci():
            replace_dev_reqs(destination_dev_req, pkg, wheel_dir)

        _inject_custom_reqs(destination_dev_req, injected_packages, pkg)

    next_proxy_port = BASE_PROXY_PORT
    for package, check in combos:
        if not is_check_enabled(package, check, CHECK_DEFAULTS.get(check, True)):
            logger.warning(f"Skipping disabled check {check} for package {package}")
            continue
        logger.info(f"Assigning proxy port {next_proxy_port} to check {check} for package {package}")

        # Check if this package overrides the Python version for analysis
        pkg_python_version = get_config_setting(package, "analyze_python_version", None)
        if pkg_python_version:
            logger.info(f"Package {package} overrides analyze Python version to {pkg_python_version}")

        scheduled.append((package, check, next_proxy_port, pkg_python_version))
        next_proxy_port += 1

    total = len(scheduled)

    for idx, (package, check, proxy_port, pkg_python_version) in enumerate(scheduled, start=1):
        tasks.append(
            asyncio.create_task(
                run_check(
                    semaphore,
                    package,
                    check,
                    base_args,
                    idx,
                    total or 1,
                    proxy_port,
                    mark_arg,
                    dest_dir,
                    service,
                    pkg_python_version,
                )
            )
        )

    # Handle Ctrl+C gracefully
    pending = set(tasks)
    try:
        results = await asyncio.gather(*tasks, return_exceptions=True)
    except KeyboardInterrupt:
        logger.warning("KeyboardInterrupt received. Cancelling running checks...")
        for t in pending:
            t.cancel()
        raise
    # Normalize exceptions
    norm_results: List[CheckResult] = []
    for res, (package, check, _, _) in zip(results, scheduled):
        if isinstance(res, CheckResult):
            norm_results.append(res)
        elif isinstance(res, Exception):
            norm_results.append(CheckResult(package, check, 99, 0.0, "", str(res)))
        else:
            norm_results.append(CheckResult(package, check, 98, 0.0, "", f"Unknown result type: {res}"))
    return summarize(norm_results)


def configure_interrupt_handling():
    """Install a SIGINT handler that triggers graceful shutdown.

    Registers a handler for SIGINT which raises :class:`KeyboardInterrupt` to allow
    the asyncio event loop to cancel tasks and subprocesses cleanly. On platforms or
    contexts where ``signal.signal`` is not supported (for example non-main threads),
    registration is skipped silently.

    :returns: None
    :rtype: None
    """

    # Ensure that a SIGINT propagates to asyncio tasks & subprocesses
    def handler(signum, frame):
        """Signal handler for SIGINT.

        Logs receipt of the signal and raises :class:`KeyboardInterrupt` to trigger
        graceful shutdown of asyncio tasks and subprocesses.

        :param signum: The numeric signal received (e.g. ``signal.SIGINT``).
        :type signum: int
        :param frame: Current stack frame when the signal was received (may be ``None``).
        :type frame: object
        :raises KeyboardInterrupt: Always raised to signal shutdown.
        """
        logger.warning(f"Received signal {signum}. Attempting graceful shutdown...")
        # Let asyncio loop raise KeyboardInterrupt
        raise KeyboardInterrupt

    try:
        signal.signal(signal.SIGINT, handler)
    except (ValueError, AttributeError):  # not supported on some platforms/threads
        pass


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="""
This script is the single point for all checks invoked by CI within this repo. It works in two phases.
    1. Identify which packages in the repo are in scope for this script invocation, based on a glob string and a service directory.
    2. Invoke one or multiple `checks` environments for each package identified as in scope.
In the case of an environment invoking `pytest`, results can be collected in a junit xml file, and test markers can be selected via --mark_arg.
"""
    )

    parser.add_argument(
        "glob_string",
        nargs="?",
        help=(
            "A comma separated list of glob strings that will target the top level directories that contain packages."
            'Examples: All = "azure-*", Single = "azure-keyvault-keys", Targeted Multiple = "azure-keyvault-keys,azure-mgmt-resource"'
        ),
    )

    parser.add_argument(
        "--junitxml",
        dest="test_results",
        help=(
            "The output path for the test results file of invoked checks."
            'Example: --junitxml="junit/test-results.xml"'
        ),
    )

    parser.add_argument(
        "--mark_arg",
        dest="mark_arg",
        help=(
            'The complete argument for `pytest -m "<input>"`. This can be used to exclude or include specific pytest markers.'
            '--mark_arg="not cosmosEmulator"'
        ),
    )

    parser.add_argument("--disablecov", help=("Flag. Disables code coverage."), action="store_true")

    parser.add_argument(
        "--service",
        help=("Name of service directory (under sdk/) to test. Example: --service applicationinsights"),
    )

    parser.add_argument(
        "-c",
        "--checks",
        dest="checks_list",
        help="Specific set of named environments to execute",
    )

    parser.add_argument(
        "-w",
        "--wheel_dir",
        dest="wheel_dir",
        help="Location for prebuilt artifacts (if any)",
    )

    parser.add_argument(
        "-i",
        "--injected-packages",
        dest="injected_packages",
        default="",
        help="Comma or space-separated list of packages that should be installed prior to dev_requirements. If local path, should be absolute.",
    )

    parser.add_argument(
        "--filter-type",
        dest="filter_type",
        default="Build",
        help="Filter type to identify eligible packages. for e.g. packages filtered in Build can pass filter type as Build,",
        choices=["Build", "Docs", "Regression", "Omit_management", "None"],
    )

    parser.add_argument(
        "-d",
        "--dest-dir",
        dest="dest_dir",
        help="Location to generate any output files (if any). For e.g. APIView stub file",
    )

    parser.add_argument(
        "--max-parallel",
        dest="max_parallel",
        type=int,
        default=os.cpu_count() or 4,
        help="Maximum number of concurrent checks (default: number of CPU cores).",
    )

    parser.add_argument(
        "--disable-compatibility-filter",
        dest="disable_compatibility_filter",
        action="store_true",
        help="Flag to disable compatibility filter while discovering packages.",
    )

    args = parser.parse_args()

    configure_logging(args)

    # We need to support both CI builds of everything and individual service
    # folders. This logic allows us to do both.
    if args.service and args.service != "auto":
        service_dir = os.path.join("sdk", args.service)
        target_dir = os.path.join(root_dir, service_dir)
    else:
        target_dir = root_dir

    logger.info(f"Beginning discovery for {args.service} and root dir {root_dir}. Resolving to {target_dir}.")

    # ensure that recursive virtual envs aren't messed with by this call
    os.environ.pop("VIRTUAL_ENV", None)
    os.environ.pop("PYTHON_HOME", None)

    if args.filter_type == "None":
        args.filter_type = "Build"
        compatibility_filter = False
    else:
        compatibility_filter = not args.disable_compatibility_filter

    targeted_packages = discover_targeted_packages(
        args.glob_string, target_dir, "", args.filter_type, compatibility_filter
    )

    if len(targeted_packages) == 0:
        logger.info(f"No packages collected for targeting string {args.glob_string} and root dir {root_dir}. Exit 0.")
        exit(0)

    logger.info(f"Executing checks with the executable {sys.executable}.")
    logger.info(f"Packages targeted: {targeted_packages}")

    temp_wheel_dir = args.wheel_dir or os.path.join(root_dir, ".wheels")
    if args.wheel_dir:
        os.environ["PREBUILT_WHEEL_DIR"] = args.wheel_dir
    else:
        if not os.path.exists(temp_wheel_dir):
            os.makedirs(temp_wheel_dir)

    if in_ci():
        build_whl_for_req("eng/tools/azure-sdk-tools", root_dir, temp_wheel_dir)

    # so if we have checks whl,import_all and selected package paths `sdk/core/azure-core`, `sdk/storage/azure-storage-blob` we should
    # shell out to `azypysdk <checkname>` with cwd of the package directory, which is what is in `targeted_packages` array
    # each individual thread may need to re-invoke if they need to self-isolate themselves, but we don't have to worry about that.

    # Prepare check list
    raw_checks = (args.checks_list or "").split(",")
    checks = [c.strip() for c in raw_checks if c and c.strip()]
    if not checks:
        logger.error("No valid checks provided via -c/--checks.")
        sys.exit(2)

    # ensure that the proxy exe is available before we start running checks that may need to populate it
    if in_ci() and _checks_require_recording_restore(checks):
        try:
            proxy_executable = prepare_local_tool(root_dir)
        except Exception as exc:
            logger.error(f"Unable to prepare test proxy executable for recording restore: {exc}")
            sys.exit(1)

    logger.info(
        f"Running {len(checks)} check(s) across {len(targeted_packages)} packages (max_parallel={args.max_parallel})."
    )

    configure_interrupt_handling()
    proxy_processes: List[ProxyProcess] = []
    try:
        if in_ci():
            logger.info(f"Ensuring {len(checks)} test proxies are running for requested checks...")
        # Pass through service if set and not "auto"
        effective_service = args.service if (args.service and args.service != "auto") else None
        exit_code = asyncio.run(
            run_all_checks(
                targeted_packages,
                checks,
                args.max_parallel,
                temp_wheel_dir,
                args.mark_arg,
                args.injected_packages,
                args.dest_dir,
                effective_service,
            )
        )
    except KeyboardInterrupt:
        logger.error("Aborted by user.")
        exit_code = 130
    finally:
        _cleanup_isolate_dirs()
    sys.exit(exit_code)



# coding=utf-8
# ------------------------------------
# Copyright (c) Microsoft.
# Licensed under the MIT License.
# ------------------------------------

"""
FILE: sample_get_model_evaluation_summary.py
DESCRIPTION:
    This sample demonstrates how to get a **model evaluation summary** for a trained model
    in a Text Authoring project (overall metrics + class metrics + confusion matrix).
USAGE:
    python sample_get_model_evaluation_summary.py
REQUIRED ENV VARS (for AAD / DefaultAzureCredential):
    AZURE_TEXT_ENDPOINT
    AZURE_CLIENT_ID
    AZURE_TENANT_ID
    AZURE_CLIENT_SECRET
NOTE:
    If you want to use AzureKeyCredential instead, set:
      - AZURE_TEXT_ENDPOINT
      - AZURE_TEXT_KEY
OPTIONAL ENV VARS:
    PROJECT_NAME         # defaults to "<project-name>"
    TRAINED_MODEL_LABEL  # defaults to "<trained-model-label>"
"""

# [START text_authoring_get_model_evaluation_summary]
import os
from azure.identity import DefaultAzureCredential
from azure.ai.textanalytics.authoring import TextAuthoringClient
from azure.ai.textanalytics.authoring.models import (
    CustomSingleLabelClassificationEvalSummary,
)


def sample_get_model_evaluation_summary():
    # settings
    endpoint = os.environ["AZURE_TEXT_ENDPOINT"]
    project_name = os.environ.get("PROJECT_NAME", "<project-name>")
    trained_model_label = os.environ.get("TRAINED_MODEL_LABEL", "<trained-model-label>")

    # create a client with AAD
    credential = DefaultAzureCredential()
    client = TextAuthoringClient(endpoint, credential=credential)

    # trained-model–scoped call
    project_client = client.get_project_client(project_name)
    eval_summary = project_client.trained_model.get_model_evaluation_summary(trained_model_label)

    print("=== Model Evaluation Summary ===")
    assert isinstance(eval_summary, CustomSingleLabelClassificationEvalSummary)
    # Evaluation options
    evaluation_options = eval_summary.evaluation_options
    print("Evaluation Options:")
    print(f"    Kind: {evaluation_options.kind}")
    print(f"    Training Split Percentage: {evaluation_options.training_split_percentage}")
    print(f"    Testing Split Percentage: {evaluation_options.testing_split_percentage}")

    # Single-label classification evaluation (micro/macro metrics)
    sl_eval = eval_summary.custom_single_label_classification_evaluation
    print(f"Micro F1: {sl_eval.micro_f1}")
    print(f"Micro Precision: {sl_eval.micro_precision}")
    print(f"Micro Recall: {sl_eval.micro_recall}")
    print(f"Macro F1: {sl_eval.macro_f1}")
    print(f"Macro Precision: {sl_eval.macro_precision}")
    print(f"Macro Recall: {sl_eval.macro_recall}")

    # Confusion matrix (dict-of-dicts with normalized/raw values)
    cmatrix = sl_eval.confusion_matrix
    if cmatrix:
        print("Confusion Matrix:")
        for row_key, row_val in cmatrix.items():
            print(f"Row: {row_key}")
            for col_key, cell in row_val.items():
                print(
                    f"    Column: {col_key}, Normalized Value: {cell['normalizedValue']}, Raw Value: {cell['rawValue']}"
                )

    # Class-specific metrics
    classes_map = sl_eval.classes
    if classes_map:
        print("Class-Specific Metrics:")
        for cls_name, metrics in classes_map.items():
            print(f"Class: {cls_name}")
            print(f"    F1: {metrics.f1}")
            print(f"    Precision: {metrics.precision}")
            print(f"    Recall: {metrics.recall}")
            print(f"    True Positives: {metrics.true_positive_count}")
            print(f"    True Negatives: {metrics.true_negative_count}")
            print(f"    False Positives: {metrics.false_positive_count}")
            print(f"    False Negatives: {metrics.false_negative_count}")


# [END text_authoring_get_model_evaluation_summary]


def main():
    sample_get_model_evaluation_summary()


if __name__ == "__main__":
    main()



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    Given an AIProjectClient, this sample demonstrates how to use the synchronous
    `.fine_tuning.jobs` methods to create reinforcement fine-tuning jobs.
    Supported OpenAI models: o4-mini

USAGE:
    python sample_finetuning_reinforcement_job.py

    Before running the sample:

    pip install "azure-ai-projects>=2.0.0" python-dotenv

    Set these environment variables with your own values:
    1) FOUNDRY_PROJECT_ENDPOINT - Required. The Azure AI Project endpoint, as found in the overview page of your
       Microsoft Foundry portal.
    2) MODEL_NAME - Optional. The base model name to use for fine-tuning. Default to the `o4-mini` model.
    3) TRAINING_FILE_PATH - Optional. Path to the training data file. Default to the `data` folder.
    4) VALIDATION_FILE_PATH - Optional. Path to the validation data file. Default to the `data` folder.
"""

import os
from typing import Any, Dict
from dotenv import load_dotenv
from fine_tuning_sample_helper import resolve_data_file_path  # pylint: disable=import-error
from azure.identity import DefaultAzureCredential
from azure.ai.projects import AIProjectClient

load_dotenv()

endpoint = os.environ["FOUNDRY_PROJECT_ENDPOINT"]
model_name = os.environ.get("MODEL_NAME", "o4-mini")
training_file_path = resolve_data_file_path(__file__, "TRAINING_FILE_PATH", "rft_training_set.jsonl")
validation_file_path = resolve_data_file_path(__file__, "VALIDATION_FILE_PATH", "rft_validation_set.jsonl")

with (
    DefaultAzureCredential() as credential,
    AIProjectClient(endpoint=endpoint, credential=credential) as project_client,
    project_client.get_openai_client() as openai_client,
):

    print("Uploading training file...")
    with open(training_file_path, "rb") as f:
        train_file = openai_client.files.create(file=f, purpose="fine-tune")
    print(f"Uploaded training file with ID: {train_file.id}")

    print("Uploading validation file...")
    with open(validation_file_path, "rb") as f:
        validation_file = openai_client.files.create(file=f, purpose="fine-tune")
    print(f"Uploaded validation file with ID: {validation_file.id}")

    print("Waits for the training and validation files to be processed...")
    openai_client.files.wait_for_processing(train_file.id)
    openai_client.files.wait_for_processing(validation_file.id)

    grader: Dict[str, Any] = {
        "name": "Response Quality Grader",
        "type": "score_model",
        "model": "o3-mini",
        "input": [
            {
                "role": "user",
                "content": "Evaluate the model's response based on correctness and quality. Rate from 0 to 10.",
            }
        ],
        "range": [0.0, 10.0],
    }

    print("Creating reinforcement fine-tuning job")
    fine_tuning_job = openai_client.fine_tuning.jobs.create(
        training_file=train_file.id,
        validation_file=validation_file.id,
        model=model_name,
        method={  # type: ignore[arg-type]
            "type": "reinforcement",
            "reinforcement": {
                "grader": grader,  # type: ignore[typeddict-item]
                "hyperparameters": {
                    "n_epochs": 1,
                    "batch_size": 4,
                    "learning_rate_multiplier": 2,
                    "eval_interval": 5,
                    "eval_samples": 2,
                    "reasoning_effort": "medium",
                },
            },
        },
        extra_body={
            "trainingType": "Standard"
        },  # Recommended approach to set trainingType. Omitting this field may lead to unsupported behavior.
    )
    print(fine_tuning_job)



# pylint: disable=line-too-long,useless-suppression
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
# pylint: disable=useless-super-delegation

import datetime
from typing import Any, List, Mapping, Optional, TYPE_CHECKING, Union, overload

from .._utils.model_base import Model as _Model, rest_field

if TYPE_CHECKING:
    from .. import models as _models


class Resource(_Model):
    """Resource.

    :ivar id: Fully qualified resource ID for the resource. Ex -
     /subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/{resourceType}/{resourceName}.
    :vartype id: str
    :ivar name: The name of the resource.
    :vartype name: str
    :ivar type: The type of the resource. E.g. "Microsoft.Compute/virtualMachines" or
     "Microsoft.Storage/storageAccounts".
    :vartype type: str
    :ivar system_data: Azure Resource Manager metadata containing createdBy and modifiedBy
     information.
    :vartype system_data: ~azure.mgmt.containerservicesafeguards.models.SystemData
    """

    id: Optional[str] = rest_field(visibility=["read"])
    """Fully qualified resource ID for the resource. Ex -
     /subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/{resourceType}/{resourceName}."""
    name: Optional[str] = rest_field(visibility=["read"])
    """The name of the resource."""
    type: Optional[str] = rest_field(visibility=["read"])
    """The type of the resource. E.g. \"Microsoft.Compute/virtualMachines\" or
     \"Microsoft.Storage/storageAccounts\"."""
    system_data: Optional["_models.SystemData"] = rest_field(name="systemData", visibility=["read"])
    """Azure Resource Manager metadata containing createdBy and modifiedBy information."""


class ExtensionResource(Resource):
    """The base extension resource.

    :ivar id: Fully qualified resource ID for the resource. Ex -
     /subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/{resourceType}/{resourceName}.
    :vartype id: str
    :ivar name: The name of the resource.
    :vartype name: str
    :ivar type: The type of the resource. E.g. "Microsoft.Compute/virtualMachines" or
     "Microsoft.Storage/storageAccounts".
    :vartype type: str
    :ivar system_data: Azure Resource Manager metadata containing createdBy and modifiedBy
     information.
    :vartype system_data: ~azure.mgmt.containerservicesafeguards.models.SystemData
    """


class DeploymentSafeguard(ExtensionResource):
    """Deployment Safeguards.

    :ivar id: Fully qualified resource ID for the resource. Ex -
     /subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/{resourceType}/{resourceName}.
    :vartype id: str
    :ivar name: The name of the resource.
    :vartype name: str
    :ivar type: The type of the resource. E.g. "Microsoft.Compute/virtualMachines" or
     "Microsoft.Storage/storageAccounts".
    :vartype type: str
    :ivar system_data: Azure Resource Manager metadata containing createdBy and modifiedBy
     information.
    :vartype system_data: ~azure.mgmt.containerservicesafeguards.models.SystemData
    :ivar properties: The resource-specific properties for this resource.
    :vartype properties:
     ~azure.mgmt.containerservicesafeguards.models.DeploymentSafeguardsProperties
    :ivar e_tag: If eTag is provided in the response body, it may also be provided as a header per
     the normal etag convention.  Entity tags are used for comparing two or more entities from the
     same requested resource. HTTP/1.1 uses entity tags in the etag (section 14.19), If-Match
     (section 14.24), If-None-Match (section 14.26), and If-Range (section 14.27) header fields.
    :vartype e_tag: str
    """

    properties: Optional["_models.DeploymentSafeguardsProperties"] = rest_field(
        visibility=["read", "create", "update", "delete", "query"]
    )
    """The resource-specific properties for this resource."""
    e_tag: Optional[str] = rest_field(name="eTag", visibility=["read"])
    """If eTag is provided in the response body, it may also be provided as a header per the normal
     etag convention.  Entity tags are used for comparing two or more entities from the same
     requested resource. HTTP/1.1 uses entity tags in the etag (section 14.19), If-Match (section
     14.24), If-None-Match (section 14.26), and If-Range (section 14.27) header fields."""

    @overload
    def __init__(
        self,
        *,
        properties: Optional["_models.DeploymentSafeguardsProperties"] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class DeploymentSafeguardsProperties(_Model):
    """DeploymentSafeguards Properties.

    :ivar provisioning_state: Provisioning State. Known values are: "Succeeded", "Failed",
     "Canceled", "Creating", "Updating", and "Deleting".
    :vartype provisioning_state: str or
     ~azure.mgmt.containerservicesafeguards.models.ProvisioningState
    :ivar level: The deployment safeguards level. Possible values are Warn and Enforce. Required.
     Known values are: "Warn" and "Enforce".
    :vartype level: str or ~azure.mgmt.containerservicesafeguards.models.DeploymentSafeguardsLevel
    :ivar excluded_namespaces: User defined list of namespaces to exclude from Deployment
     Safeguards. Deployments in these namespaces will not be checked against any safeguards.
    :vartype excluded_namespaces: list[str]
    :ivar system_excluded_namespaces: System defined list of namespaces excluded from Deployment
     Safeguards. These are determined by the underlying provider (such as AKS), and cannot be
     changed. Deployments in these namespaces will not be checked. Required.
    :vartype system_excluded_namespaces: list[str]
    :ivar pod_security_standards_level: The pod security standards level. Known values are:
     "Privileged", "Baseline", and "Restricted".
    :vartype pod_security_standards_level: str or
     ~azure.mgmt.containerservicesafeguards.models.PodSecurityStandardsLevel
    """

    provisioning_state: Optional[Union[str, "_models.ProvisioningState"]] = rest_field(
        name="provisioningState", visibility=["read"]
    )
    """Provisioning State. Known values are: \"Succeeded\", \"Failed\", \"Canceled\", \"Creating\",
     \"Updating\", and \"Deleting\"."""
    level: Union[str, "_models.DeploymentSafeguardsLevel"] = rest_field(
        visibility=["read", "create", "update", "delete", "query"]
    )
    """The deployment safeguards level. Possible values are Warn and Enforce. Required. Known values
     are: \"Warn\" and \"Enforce\"."""
    excluded_namespaces: Optional[List[str]] = rest_field(
        name="excludedNamespaces", visibility=["read", "create", "update", "delete", "query"]
    )
    """User defined list of namespaces to exclude from Deployment Safeguards. Deployments in these
     namespaces will not be checked against any safeguards."""
    system_excluded_namespaces: List[str] = rest_field(name="systemExcludedNamespaces", visibility=["read"])
    """System defined list of namespaces excluded from Deployment Safeguards. These are determined by
     the underlying provider (such as AKS), and cannot be changed. Deployments in these namespaces
     will not be checked. Required."""
    pod_security_standards_level: Optional[Union[str, "_models.PodSecurityStandardsLevel"]] = rest_field(
        name="podSecurityStandardsLevel", visibility=["read", "create", "update", "delete", "query"]
    )
    """The pod security standards level. Known values are: \"Privileged\", \"Baseline\", and
     \"Restricted\"."""

    @overload
    def __init__(
        self,
        *,
        level: Union[str, "_models.DeploymentSafeguardsLevel"],
        excluded_namespaces: Optional[List[str]] = None,
        pod_security_standards_level: Optional[Union[str, "_models.PodSecurityStandardsLevel"]] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class ErrorAdditionalInfo(_Model):
    """The resource management error additional info.

    :ivar type: The additional info type.
    :vartype type: str
    :ivar info: The additional info.
    :vartype info: any
    """

    type: Optional[str] = rest_field(visibility=["read"])
    """The additional info type."""
    info: Optional[Any] = rest_field(visibility=["read"])
    """The additional info."""


class ErrorDetail(_Model):
    """The error detail.

    :ivar code: The error code.
    :vartype code: str
    :ivar message: The error message.
    :vartype message: str
    :ivar target: The error target.
    :vartype target: str
    :ivar details: The error details.
    :vartype details: list[~azure.mgmt.containerservicesafeguards.models.ErrorDetail]
    :ivar additional_info: The error additional info.
    :vartype additional_info:
     list[~azure.mgmt.containerservicesafeguards.models.ErrorAdditionalInfo]
    """

    code: Optional[str] = rest_field(visibility=["read"])
    """The error code."""
    message: Optional[str] = rest_field(visibility=["read"])
    """The error message."""
    target: Optional[str] = rest_field(visibility=["read"])
    """The error target."""
    details: Optional[List["_models.ErrorDetail"]] = rest_field(visibility=["read"])
    """The error details."""
    additional_info: Optional[List["_models.ErrorAdditionalInfo"]] = rest_field(
        name="additionalInfo", visibility=["read"]
    )
    """The error additional info."""


class ErrorResponse(_Model):
    """Error response.

    :ivar error: The error object.
    :vartype error: ~azure.mgmt.containerservicesafeguards.models.ErrorDetail
    """

    error: Optional["_models.ErrorDetail"] = rest_field(visibility=["read", "create", "update", "delete", "query"])
    """The error object."""

    @overload
    def __init__(
        self,
        *,
        error: Optional["_models.ErrorDetail"] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Operation(_Model):
    """REST API Operation.

    :ivar name: The name of the operation, as per Resource-Based Access Control (RBAC). Examples:
     "Microsoft.Compute/virtualMachines/write", "Microsoft.Compute/virtualMachines/capture/action".
    :vartype name: str
    :ivar is_data_action: Whether the operation applies to data-plane. This is "true" for
     data-plane operations and "false" for Azure Resource Manager/control-plane operations.
    :vartype is_data_action: bool
    :ivar display: Localized display information for this particular operation.
    :vartype display: ~azure.mgmt.containerservicesafeguards.models.OperationDisplay
    :ivar origin: The intended executor of the operation; as in Resource Based Access Control
     (RBAC) and audit logs UX. Default value is "user,system". Known values are: "user", "system",
     and "user,system".
    :vartype origin: str or ~azure.mgmt.containerservicesafeguards.models.Origin
    :ivar action_type: Extensible enum. Indicates the action type. "Internal" refers to actions
     that are for internal only APIs. "Internal"
    :vartype action_type: str or ~azure.mgmt.containerservicesafeguards.models.ActionType
    """

    name: Optional[str] = rest_field(visibility=["read"])
    """The name of the operation, as per Resource-Based Access Control (RBAC). Examples:
     \"Microsoft.Compute/virtualMachines/write\",
     \"Microsoft.Compute/virtualMachines/capture/action\"."""
    is_data_action: Optional[bool] = rest_field(name="isDataAction", visibility=["read"])
    """Whether the operation applies to data-plane. This is \"true\" for data-plane operations and
     \"false\" for Azure Resource Manager/control-plane operations."""
    display: Optional["_models.OperationDisplay"] = rest_field(
        visibility=["read", "create", "update", "delete", "query"]
    )
    """Localized display information for this particular operation."""
    origin: Optional[Union[str, "_models.Origin"]] = rest_field(visibility=["read"])
    """The intended executor of the operation; as in Resource Based Access Control (RBAC) and audit
     logs UX. Default value is \"user,system\". Known values are: \"user\", \"system\", and
     \"user,system\"."""
    action_type: Optional[Union[str, "_models.ActionType"]] = rest_field(name="actionType", visibility=["read"])
    """Extensible enum. Indicates the action type. \"Internal\" refers to actions that are for
     internal only APIs. \"Internal\""""

    @overload
    def __init__(
        self,
        *,
        display: Optional["_models.OperationDisplay"] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class OperationDisplay(_Model):
    """Localized display information for and operation.

    :ivar provider: The localized friendly form of the resource provider name, e.g. "Microsoft
     Monitoring Insights" or "Microsoft Compute".
    :vartype provider: str
    :ivar resource: The localized friendly name of the resource type related to this operation.
     E.g. "Virtual Machines" or "Job Schedule Collections".
    :vartype resource: str
    :ivar operation: The concise, localized friendly name for the operation; suitable for
     dropdowns. E.g. "Create or Update Virtual Machine", "Restart Virtual Machine".
    :vartype operation: str
    :ivar description: The short, localized friendly description of the operation; suitable for
     tool tips and detailed views.
    :vartype description: str
    """

    provider: Optional[str] = rest_field(visibility=["read"])
    """The localized friendly form of the resource provider name, e.g. \"Microsoft Monitoring
     Insights\" or \"Microsoft Compute\"."""
    resource: Optional[str] = rest_field(visibility=["read"])
    """The localized friendly name of the resource type related to this operation. E.g. \"Virtual
     Machines\" or \"Job Schedule Collections\"."""
    operation: Optional[str] = rest_field(visibility=["read"])
    """The concise, localized friendly name for the operation; suitable for dropdowns. E.g. \"Create
     or Update Virtual Machine\", \"Restart Virtual Machine\"."""
    description: Optional[str] = rest_field(visibility=["read"])
    """The short, localized friendly description of the operation; suitable for tool tips and detailed
     views."""


class SystemData(_Model):
    """Metadata pertaining to creation and last modification of the resource.

    :ivar created_by: The identity that created the resource.
    :vartype created_by: str
    :ivar created_by_type: The type of identity that created the resource. Known values are:
     "User", "Application", "ManagedIdentity", and "Key".
    :vartype created_by_type: str or ~azure.mgmt.containerservicesafeguards.models.CreatedByType
    :ivar created_at: The timestamp of resource creation (UTC).
    :vartype created_at: ~datetime.datetime
    :ivar last_modified_by: The identity that last modified the resource.
    :vartype last_modified_by: str
    :ivar last_modified_by_type: The type of identity that last modified the resource. Known values
     are: "User", "Application", "ManagedIdentity", and "Key".
    :vartype last_modified_by_type: str or
     ~azure.mgmt.containerservicesafeguards.models.CreatedByType
    :ivar last_modified_at: The timestamp of resource last modification (UTC).
    :vartype last_modified_at: ~datetime.datetime
    """

    created_by: Optional[str] = rest_field(name="createdBy", visibility=["read", "create", "update", "delete", "query"])
    """The identity that created the resource."""
    created_by_type: Optional[Union[str, "_models.CreatedByType"]] = rest_field(
        name="createdByType", visibility=["read", "create", "update", "delete", "query"]
    )
    """The type of identity that created the resource. Known values are: \"User\", \"Application\",
     \"ManagedIdentity\", and \"Key\"."""
    created_at: Optional[datetime.datetime] = rest_field(
        name="createdAt", visibility=["read", "create", "update", "delete", "query"], format="rfc3339"
    )
    """The timestamp of resource creation (UTC)."""
    last_modified_by: Optional[str] = rest_field(
        name="lastModifiedBy", visibility=["read", "create", "update", "delete", "query"]
    )
    """The identity that last modified the resource."""
    last_modified_by_type: Optional[Union[str, "_models.CreatedByType"]] = rest_field(
        name="lastModifiedByType", visibility=["read", "create", "update", "delete", "query"]
    )
    """The type of identity that last modified the resource. Known values are: \"User\",
     \"Application\", \"ManagedIdentity\", and \"Key\"."""
    last_modified_at: Optional[datetime.datetime] = rest_field(
        name="lastModifiedAt", visibility=["read", "create", "update", "delete", "query"], format="rfc3339"
    )
    """The timestamp of resource last modification (UTC)."""

    @overload
    def __init__(
        self,
        *,
        created_by: Optional[str] = None,
        created_by_type: Optional[Union[str, "_models.CreatedByType"]] = None,
        created_at: Optional[datetime.datetime] = None,
        last_modified_by: Optional[str] = None,
        last_modified_by_type: Optional[Union[str, "_models.CreatedByType"]] = None,
        last_modified_at: Optional[datetime.datetime] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    This sample demonstrates how to use Agent operations with a mix of automatic and manual
    function calls from the Azure Agents service using a synchronous client.

    Some functions (like fetch_current_datetime and fetch_weather) will be executed automatically,
    while others (like send_email) will require manual approval through the RunHandler.

USAGE:
    python sample_agents_functions_in_create_and_process.py

    Before running the sample:

    pip install azure-ai-projects azure-ai-agents azure-identity

    Set these environment variables with your own values:
    1) PROJECT_ENDPOINT - The Azure AI Project endpoint, as found in the Overview
                          page of your Azure AI Foundry portal.
    2) MODEL_DEPLOYMENT_NAME - The deployment name of the AI model, as found under the "Name" column in
       the "Models + endpoints" tab in your Azure AI Foundry project.
"""
import os, sys, json
from typing import Any, Optional
from azure.ai.projects import AIProjectClient
from azure.identity import DefaultAzureCredential
from azure.ai.agents.models import (
    FunctionTool,
    ListSortOrder,
    RequiredFunctionToolCall,
    RunHandler,
    ThreadRun,
    ToolSet,
    RequiredFunctionToolCallDetails,
)


# Add package directory to sys.path to import user_functions
current_dir = os.path.dirname(os.path.abspath(__file__))
package_dir = os.path.abspath(os.path.join(current_dir, os.pardir, os.pardir))
if package_dir not in sys.path:
    sys.path.insert(0, package_dir)
from samples.utils.user_functions import (
    fetch_current_datetime,
    fetch_weather,
    send_email,
    calculate_sum,
)
from samples.utils.user_functions import user_functions

project_client = AIProjectClient(
    endpoint=os.environ["PROJECT_ENDPOINT"],
    credential=DefaultAzureCredential(),
)

# Create a toolset with all functions
all_functions_tool = FunctionTool(user_functions)
toolset = ToolSet()
toolset.add(all_functions_tool)


# Declare a run handler to execute function tool call manually
# [START run_handler]
class MyRunHandler(RunHandler):
    def submit_function_call_output(
        self,
        *,
        run: ThreadRun,
        tool_call: RequiredFunctionToolCall,
        tool_call_details: RequiredFunctionToolCallDetails,
        **kwargs: Any,
    ) -> Any:
        function_name = tool_call_details.name
        if function_name == send_email.__name__:
            # Parse arguments from tool call
            args_dict = json.loads(tool_call_details.arguments) if tool_call_details.arguments else {}
            # Call the function directly with the arguments
            return send_email(**args_dict)


# [END run_handler]

with project_client:
    agents_client = project_client.agents

    # Enable auto function calling for the subset of safe functions
    auto_functions = {
        fetch_current_datetime,
        fetch_weather,
        calculate_sum,
    }
    agents_client.enable_auto_function_calls(auto_functions)

    # Create an Agent with all function tools available
    agent = agents_client.create_agent(
        model=os.environ["MODEL_DEPLOYMENT_NAME"],
        name="my-agent",
        instructions="You are a helpful Agent",
        toolset=toolset,
    )
    print(f"Created Agent, ID: {agent.id}")

    thread = agents_client.threads.create()
    print(f"Created thread, ID: {thread.id}")

    message = agents_client.messages.create(
        thread_id=thread.id,
        role="user",
        content="Hello, send an email with the datetime and weather information in New York?",
    )
    print(f"Created message, ID: {message.id}")

    # [START create_and_process]
    run = agents_client.runs.create_and_process(thread_id=thread.id, agent_id=agent.id, run_handler=MyRunHandler())
    # [END create_and_process]

    print(f"Run completed with status: {run.status}")

    # Delete the Agent when done
    agents_client.delete_agent(agent.id)
    print("Deleted Agent")

    # Fetch and log all messages
    messages = agents_client.messages.list(thread_id=thread.id, order=ListSortOrder.ASCENDING)
    for msg in messages:
        if msg.text_messages:
            last_text = msg.text_messages[-1]
            print(f"{msg.role}: {last_text.text.value}")



import argparse

from typing import Optional, List

from .pylint import pylint


class next_pylint(pylint):
    def __init__(self):
        super().__init__()

    def register(
        self, subparsers: "argparse._SubParsersAction", parent_parsers: Optional[List[argparse.ArgumentParser]] = None
    ) -> None:
        """Register the next-pylint check. The next-pylint check installs the next version of pylint and runs pylint against the target package."""
        parents = parent_parsers or []
        p = subparsers.add_parser(
            "next-pylint", parents=parents, help="Run the pylint check with the next version of pylint"
        )
        p.set_defaults(func=self.run)

    def run(self, args: argparse.Namespace) -> int:
        """Run the next-pylint check command."""
        args.next = True
        return super().run(args)
